/*
  COPYRIGHT NOTICE
  Copyright (c) 2009  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Array;
import java.util.Vector;

/**
 * Processor that searches for an object in an object array.
 *
 * @version 3 February 2009
 */
public class ArraySearcher extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Searches for an object in an array.";

    /** Input Node name for the object to look for. */
    private static String OBJ_NAME = "Object";

    /** Input Node description for the object to look for. */
    private static String OBJ_DESC = "Object to search for.";

    /** Input Node name for the array to look in. */
    private static String ARRAY_NAME = "Array";

    /** Input Node description for the array to look in. */
    private static String ARRAY_DESC = "Array to search.";

    /** Output Node name for whether the object is in the array. */
    private static String IN_NAME = "In Array?";

    /** Output Node description for whether the object is in the array. */
    private static String IN_DESC = "Is the object in the array?";

    /** Property name for the object type. */
    private static String PROP_TYPE_NAME = "type";

    /** Property description for the object type. */
    private static String PROP_TYPE_DESC = "Type of object.";

    /**
     * Constructs an Array Searcher with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ArraySearcher(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs an Array Searcher using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ArraySearcher(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Get the object type
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);

	// Create the Input Node for the array
	Class arrayType = Array.newInstance(type, 0).getClass();
	_addInputNode( new InputNode(ARRAY_NAME, ARRAY_DESC, arrayType) );

	// Create the Input Node for the object
	_addInputNode( new InputNode(OBJ_NAME, OBJ_DESC, type) );
      
	// Create the Output Node
	type = Boolean.class;
	_addOutputNode( new OutputNode(IN_NAME, IN_DESC, type) );
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the array
	    Object array = null;
	    Object arrayValue = _getInputValue(ARRAY_NAME, null);
	    if (arrayValue != null && arrayValue.getClass().isArray()) {
		array = arrayValue;
	    }

	    // Get the object
	    Object value = _getInputValue(OBJ_NAME, null);

	    // Look for the object in the array
	    if (array != null && value != null) {
		int length = Array.getLength(array);
		for (int i = 0; i < length; i++) {
		    Object elem = Array.get(array, i);
		    if ( value.equals(elem) ) {
			_setOutputValue(IN_NAME, Boolean.TRUE);
			return;
		    }
		}
	    }

	    // Undetermined
	    _setOutputValue(IN_NAME, Boolean.FALSE);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[1];

	// Primitive classes
	Vector optionSet = new Vector();
	optionSet.add( Boolean.class );
	optionSet.add( Byte.class );
	optionSet.add( Double.class );
	optionSet.add( Float.class );
	optionSet.add( Integer.class );
	optionSet.add( Long.class );
	optionSet.add( Short.class );
	optionSet.add( String.class );

	// Create a Set Restriction
	String name = "Allowed Object Types";
	String desc = "Type of the objects in the array.";
	Restriction restriction = new SetRestriction(name, desc, optionSet);

	// Default type
	name = PROP_TYPE_NAME;
	desc = PROP_TYPE_DESC;
	Object value = String.class;
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
