/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Constructor;
import java.util.Vector;

/**
 * Processor that produces a constant output value.  If the value is a String,
 * new line characters are encoded as '\n' characters.
 *
 * @version 5 June 2003
 */
public class Constant extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Produces a constant value.";

  /** Name of the Output Node. */
  private static String NODE_NAME = "Constant";

  /** Property for the value type. */
  private static String PROP_TYPE_NAME = "type";

  /** Property description for the value type. */
  private static String PROP_TYPE_DESC = "Type of constant.";

  /** Property for the value. */
  private static String PROP_VALUE_NAME = "value";

  /** Property description for the value. */
  private static String PROP_VALUE_DESC = "Constant value.";

  /** Property for the value description. */
  private static String PROP_DESC_NAME = "desc";

  /** Property description for the value description. */
  private static String PROP_DESC_DESC = "Description of the constant.";

  /**
   * Constructs a Constant with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Constant(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Constant using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Constant(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the constant type and description
      Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
      String desc = (String)_getPropertyValue(PROP_DESC_NAME);

      // Create an Output Node for the constant value
      OutputNode outputNode = new OutputNode(NODE_NAME, desc, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      // Set the constant value
      try {

	// Get the constant type and value
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
	String value = (String)_getPropertyValue(PROP_VALUE_NAME);

	// Get the Constructor of the Class with one String argument
	Class[] parameters = new Class[1];
	parameters[0] = String.class;
	Constructor constructor = type.getDeclaredConstructor(parameters);

	// Set a new instance of the Class
	Object[] arguments = new Object[1];
	arguments[0] = value;
	Object constant = constructor.newInstance(arguments);
	_setOutputValue(NODE_NAME, constant);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Primitive classes
      Vector optionSet = new Vector();
      optionSet.add( Boolean.class );
      optionSet.add( Byte.class );
      optionSet.add( Double.class );
      optionSet.add( Float.class );
      optionSet.add( Integer.class );
      optionSet.add( Long.class );
      optionSet.add( Short.class );
      optionSet.add( String.class );

      // Create a Set Restriction
      String name = "Allowed Primitive Types";
      String desc = "Types that can be converted to.";
      Restriction restriction = new SetRestriction(name, desc, optionSet);

      // Default value type
      name = PROP_TYPE_NAME;
      desc = PROP_TYPE_DESC;
      Object value = Integer.class;
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default value
      name = PROP_VALUE_NAME;
      desc = PROP_VALUE_DESC;
      value = "0";
      properties[1] = new ProcessorProperty(name, desc, value);

      // Default value description
      name = PROP_DESC_NAME;
      desc = PROP_DESC_DESC;
      value = "";
      properties[2] = new ProcessorProperty(name, desc, value);

      return properties;
    }
}
