/*
  COPYRIGHT NOTICE
  Copyright (c) 2008  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that computes a hash code for a string.
 *
 * @version 10 January 2008
 */
public class StringHasher extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Computes a hash code for a string.";

    /** Input Node name for the input string. */
    private static String INPUT_NAME = "String";

    /** Input Node description for the input string. */
    private static String INPUT_DESC = "String to hash.";

    /** Output Node name for the hash code. */
    private static String HASH_NAME = "Hash Code";

    /** Output Node description for the hash code. */
    private static String HASH_DESC = "Hash code computed from the string.";

    /**
     * Constructs a String Hasher with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public StringHasher(String name)
    {
	super(name, DESC);

	// Create the Input Node
	Class type = String.class;
	_addInputNode( new InputNode(INPUT_NAME, INPUT_DESC, type) );

	// Create the Output Node
	type = Long.class;
	_addOutputNode( new OutputNode(HASH_NAME, HASH_DESC, type) );
    }

    /**
     * Constructs a String Hasher using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public StringHasher(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Gets the long description.
     *
     * @return A verbose description of what the Processor does.
     */
    public String getLongDescription()
    {
	StringBuffer b = new StringBuffer();

	b.append("This processor maps an input string to a number.  This\n");
	b.append("is done by computing a hash code for the string using:\n");
	b.append("\n");
	b.append("  hash = s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]\n");
	b.append("\n");
	b.append("  where s[i] is the ith character of the string\n");
	b.append("  and long arithmetic is used.\n");
	b.append("\n");

	return b.toString();
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the input string
	    String s = (String)_getInputValue(INPUT_NAME, null);
	    if (s == null) { return; }

	    // Compute the hash code (from Java's String source code)
	    long hashCode = 0;
	    for (int i = 0; i < s.length(); i++) {
		hashCode = 31*hashCode + s.charAt(i);
	    }

	    // No negatives
	    if (hashCode < 0) { hashCode *= -1; }

	    // Set the output
	    _setOutputValue(HASH_NAME, new Long(hashCode));
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }
}
