/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that truncates the end of a String.
 *
 * @version 18 June 2003
 */
public class StringTruncator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Truncates the end of a String.";

  /** Input Node Name for the String to truncate. */
  private static String STRING_NAME = "String";

  /** Input Node Description for the String to truncate. */
  private static String STRING_DESC = "String to truncate.";

  /** Output Node Name for the truncated String. */
  private static String TRUNCATED_NAME = "Truncated String";

  /** Output Node Description for the truncated String. */
  private static String TRUNCATED_DESC = "String whose end is truncated.";

  /** Property for the maximum length of the String. */
  private static String PROP_MAX_NAME = "max";

  /** Property description for the maximum length of the String. */
  private static String PROP_MAX_DESC = "Length to limit the string by.";

  /**
   * Constructs a String Truncator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringTruncator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a String Truncator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringTruncator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = String.class;
      OutputNode node = new OutputNode(TRUNCATED_NAME, TRUNCATED_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the input value
	String s = (String)_getInputValue(STRING_NAME, "");

	// Get the maximum length
	Integer max = (Integer)_getPropertyValue(PROP_MAX_NAME);

	// Truncate the String
	String truncatedS = null;
	if (s != null) {
	  truncatedS = s;
	  
	  // Only truncate the String if it is too big
	  if ( truncatedS.length() > max.intValue() ) {
	    truncatedS = truncatedS.substring(0, max.intValue());
	  }
	}

	// Set the truncated String
	_setOutputValue(TRUNCATED_NAME, truncatedS);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default substring
      String name = PROP_MAX_NAME;
      String desc = PROP_MAX_DESC;
      Object value = new Integer(0);
      properties[0] = new ProcessorProperty(name, desc, value);

      return properties;
    }
}
