/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.HashMap;

/**
 * Processor that maps keys to values.
 *
 * @version 30 June 2003
 */
public class ValueMapper extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Maps keys to values.";

  /** Input Node name for the actual key. */
  private static String KEY_NAME = "Key";

  /** Input Node description for the actual key. */
  private static String KEY_DESC = "Key to find a value for.";

  /** Base Input Node name for the ith key. */
  private static String KEY_BASE_NAME = "Key #";

  /** Base Input Node description for a key. */
  private static String KEY_BASE_DESC = "Key that maps to value #";

  /** Base Input Node name for the ith value. */
  private static String VALUE_BASE_NAME = "Value #";

  /** Base Input Node description for a value. */
  private static String VALUE_BASE_DESC = "Value mapped from key #";

  /** Output Node name for the actual value. */
  private static String VALUE_NAME = "Value";

  /** Output Node description for the actual value. */
  private static String VALUE_DESC = "Value mapped from the key.";

  /** Property name for the number of key-value mappings. */
  private static String PROP_NUMBER_NAME = "number";

  /** Property description for the number of key-value mappings. */
  private static String PROP_NUMBER_DESC = "Number of key-value mappings.";

  /** Property name for the key type. */
  private static String PROP_KEY_TYPE_NAME = "key type";

  /** Property description for the key type. */
  private static String PROP_KEY_TYPE_DESC = "Type of keys.";

  /** Property name for the value type. */
  private static String PROP_VALUE_TYPE_NAME = "value type";

  /** Property description for the value type. */
  private static String PROP_VALUE_TYPE_DESC = "Type of values.";

  /**
   * Constructs a Value Mapper with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ValueMapper(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Value Mapper using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ValueMapper(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the number of key-value mappings and key & value types
      int number = ((Integer)_getPropertyValue(PROP_NUMBER_NAME)).intValue();
      Class keyType = (Class)_getPropertyValue(PROP_KEY_TYPE_NAME);
      Class valueType = (Class)_getPropertyValue(PROP_VALUE_TYPE_NAME);

      // Create the Input and Output Node for the actual key and value
      InputNode iNode = new InputNode(KEY_NAME, KEY_DESC, keyType);
      _addInputNode(iNode);

      OutputNode oNode = new OutputNode(VALUE_NAME, VALUE_DESC, valueType);
      _addOutputNode(oNode);

      // Create the Input Nodes for the key-value mappings
      for (int i = 0; i < number; i++) {
	String iName = KEY_BASE_NAME + i;
	String iDesc = KEY_BASE_DESC + i;
	_addInputNode( new InputNode(iName, iDesc,  keyType) );

	iName = VALUE_BASE_NAME + i;
	iDesc = VALUE_BASE_DESC + i;
	_addInputNode( new InputNode(iName, iDesc,  valueType) );
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the number of key-value mappings
	int number = ((Integer)_getPropertyValue(PROP_NUMBER_NAME)).intValue();

	// Create a Hash Map using the Input Node key-value mappings
	HashMap hashMap = new HashMap(number);
	for (int i = 0; i < number; i++) {
	  Object key = _getInputValue(KEY_BASE_NAME + i, null);
	  Object value = _getInputValue(VALUE_BASE_NAME + i, null);

	  hashMap.put(key, value);
	}

	// Set the Output Node value using the key
	Object key = _getInputValue(KEY_NAME, null);
	_setOutputValue(VALUE_NAME, hashMap.get(key));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Default number of key-value mappings
      String name = PROP_NUMBER_NAME;
      String desc = PROP_NUMBER_DESC;
      Object value = new Integer(1);
      Restriction restriction = Utilities.getPositiveRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default key type
      name = PROP_KEY_TYPE_NAME;
      desc = PROP_KEY_TYPE_DESC;
      value = String.class;
      restriction = Utilities.getClassRestriction();
      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      // Default value type
      name = PROP_VALUE_TYPE_NAME;
      desc = PROP_VALUE_TYPE_DESC;
      value = String.class;
      restriction = Utilities.getClassRestriction();
      properties[2] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}



