/*
  COPYRIGHT NOTICE
  Copyright (c) 2007  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that parses a DICOM time string.
 *
 * @version 12 January 2007
 */
public class DicomTimeParser extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Parses a DICOM time string.";

  /** Input Node name for the DICOM time string. */
  private static String TIME_NAME = "DICOM Time String";

  /** Input Node description for the DICOM time string. */
  private static String TIME_DESC = "String in the format hhmmss.frac.";

  /** Output Node name for the parsed hour. */
  private static String HOUR_NAME = "Hour";

  /** Output Node description for the parsed hour. */
  private static String HOUR_DESC = "Hour (0-23) parsed from the time string.";

  /** Output Node name for the parsed minutes. */
  private static String MINUTE_NAME = "Minutes";

  /** Output Node description for the parsed minutes. */
  private static String MINUTE_DESC = 
  "Minutes (0-59) parsed from the time string.";

  /** Output Node name for the parsed seconds. */
  private static String SECOND_NAME = "Seconds";

  /** Output Node description for the parsed seconds. */
  private static String SECOND_DESC =
  "Seconds (0-59) parsed from the time string.";

  /** Output Node name for the parsed milliseconds. */
  private static String MILLI_NAME = "Milliseconds";

  /** Output Node description for the parsed milliseconds. */
  private static String MILLI_DESC =
  "Milliseconds (0-999) parsed from the time string.";

  /** Output Node name for the parsed time. */
  private static String TOTAL_NAME = "Time";

  /** Output Node description for the parsed time. */
  private static String TOTAL_DESC = "Total number of milliseconds.";

  /** Property name for choosing the year/month/day or time. */
  private static String PROP_NAME = "Output Format";

  /** Property description for choosing the year/month/day or time. */
  private static String PROP_DESC = "Format for the parsed output.";

  /**
   * Constructs a DICOM Time Parser with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTimeParser(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a DICOM Time Parser using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTimeParser(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the output format
      String format = (String)_getPropertyValue(PROP_NAME);

      // Create the Input Node
      Class type = String.class;
      _addInputNode( new InputNode(TIME_NAME, TIME_DESC, type) );

      // Create the Output Nodes
      if ( format.equals("Hour/Minutes/Seconds/Milli") ) {
	type = Integer.class;
	_addOutputNode( new OutputNode(HOUR_NAME, HOUR_DESC, type) );
	_addOutputNode( new OutputNode(MINUTE_NAME, MINUTE_DESC, type) );
	_addOutputNode( new OutputNode(SECOND_NAME, SECOND_DESC, type) );
	_addOutputNode( new OutputNode(MILLI_NAME, MILLI_DESC, type) );
      }
      else {
	type = Long.class;
	_addOutputNode( new OutputNode(TOTAL_NAME, TOTAL_DESC, type) );
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the time string
	String time = (String)_getInputValue(TIME_NAME, "");
	time = time.trim();
	if ( time.length() == 0 ) { return; }

	// Old versions of DICOM support hh:mm:ss.frac -> remove all :'s
	time = time.replaceAll(":", "");

	// Parse the hour, minutes, seconds, and milliseconds
	int hour = Integer.parseInt( time.substring(0, 2) );
	int minutes = 0;
	if (time.length() >= 4) {
	  minutes = Integer.parseInt( time.substring(2, 4) );
	}
	int seconds = 0;
	if (time.length() >= 6) {
	  seconds = Integer.parseInt( time.substring(4, 6) );
	}
	int millis = 0;
	if (time.length() >= 8) {

	  // Check for a "."
	  if ( time.substring(6, 7).equals(".") ) {
	    String frac = "0" + time.substring(6, time.length());
	    double dFrac = Double.parseDouble(frac);
	    millis = (int)Math.rint(1000*dFrac);
	  }
	}
	  
	// Get the output format
	String format = (String)_getPropertyValue(PROP_NAME);

	// Set the Output Nodes
	if ( format.equals("Hour/Minutes/Seconds/Milli") ) {
	  _setOutputValue(HOUR_NAME, new Integer(hour));
	  _setOutputValue(MINUTE_NAME, new Integer(minutes));
	  _setOutputValue(SECOND_NAME, new Integer(seconds));
	  _setOutputValue(MILLI_NAME, new Integer(millis));
	}

	else {
	  long t = hour*60*60*1000 + minutes*60*1000 + seconds*1000 + millis;
	  _setOutputValue(TOTAL_NAME, new Long(t));
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Create the Restriction set
      Vector optionSet = new Vector();
      optionSet.add("Hour/Minutes/Seconds/Milli");
      optionSet.add("Time");

      // Create the Restriction
      String name = PROP_NAME;
      String desc = PROP_DESC;
      SetRestriction r = new SetRestriction(name, desc, optionSet);

      // Return the Property
      properties[0] = new ProcessorProperty(name, desc,
					    "Hour/Minutes/Seconds/Milli", r);
      return properties;
    }
}
