/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that converts from DICOM space to NIFTI space.
 *
 * @version 3 November 2003
 */
public class DicomToNiftiSpace extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Converts from DICOM space to NIFTI space.";

  /** Input Node name for the X coordinate of the row direction cosine. */
  private static String X_ROW_NAME = "Row Direction Cosine (X)";

  /** Input Node description for the X coord of the row direction cosine. */
  private static String X_ROW_DESC = "X coord of the row direction cosine.";

  /** Input Node name for the Y coordinate of the row direction cosine. */
  private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

  /** Input Node description for the Y coord of the row direction cosine. */
  private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

  /** Input Node name for the Z coordinate of the row direction cosine. */
  private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

  /** Input Node description for the Z coord of the row direction cosine. */
  private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

  /** Input Node name for the X coordinate of the column direction cosine. */
  private static String X_COL_NAME = "Column Direction Cosine (X)";

  /** Input Node description for X coord of the column direction cosine. */
  private static String X_COL_DESC = "X coord of the column direction cosine.";

  /** Input Node name for the Y coordinate of the column direction cosine. */
  private static String Y_COL_NAME = "Column Direction Cosine (Y)";

  /** Input Node description for Y coord of the column direction cosine. */
  private static String Y_COL_DESC = "Y coord of the column direction cosine.";

  /** Input Node name for the Z coordinate of the column direction cosine. */
  private static String Z_COL_NAME = "Column Direction Cosine (Z)";

  /** Input Node description for Z coord of the column direction cosine. */
  private static String Z_COL_DESC = "Z coord of the column direction cosine.";

  /** Input Node name for the X coordinate of the stack direction cosine. */
  private static String X_STK_NAME = "Stack Direction Cosine (X)";

  /** Input Node description for X coord of the stack direction cosine. */
  private static String X_STK_DESC = "X coord of the stack direction cosine.";

  /** Input Node name for the Y coordinate of the stack direction cosine. */
  private static String Y_STK_NAME = "Stack Direction Cosine (Y)";

  /** Input Node description for Y coord of the stack direction cosine. */
  private static String Y_STK_DESC = "Y coord of the stack direction cosine.";

  /** Input Node name for the Z coordinate of the stack direction cosine. */
  private static String Z_STK_NAME = "Stack Direction Cosine (Z)";

  /** Input Node description for Z coord of the stack direction cosine. */
  private static String Z_STK_DESC = "Z coord of the stack direction cosine.";

  /** Input Node name for the X coord of the image position. */
  private static String X_IMG_NAME = "Image Position (Patient) X";

  /** Input Node description for the X coord of the image position. */
  private static String X_IMG_DESC = "X coord of the image position.";

  /** Input Node name for the Y coord of the image position. */
  private static String Y_IMG_NAME = "Image Position (Patient) Y";

  /** Input Node description for the Y coord of the image position. */
  private static String Y_IMG_DESC = "Y coord of the image position.";

  /** Input Node name for the Z coord of the image position. */
  private static String Z_IMG_NAME = "Image Position (Patient) Z";

  /** Input Node description for the Z coord of the image position. */
  private static String Z_IMG_DESC = "Z coord of the image position.";

  /** Output Node name for the code that specifies the coordinate system. */
  private static String QFORM_CODE_NAME = "Q Form Code";

  /** Output Node description for the code that specifies the coord system. */
  private static String QFORM_CODE_DESC =
  "Gives the method by which continuous coordinates are attached to voxels.";

  /** Output Node name for the scaling factor used in the rotation matrix. */
  private static String QFAC_NAME = "Q Factor";

  /** Output Node description for the scaling factor used in the rot matrix. */
  private static String QFAC_DESC =
  "Scaling factor used in the rotation matrix.";

  /** Output Node name for the quaternion b parameter. */
  private static String QUATERN_B_NAME = "Quaternion b Parameter";

  /** Output Node description for the quaternion b parameter. */
  private static String QUATERN_B_DESC =
  "B parameter used in the unit quaternion (a,b,c,d) to represent a rotation.";

  /** Output Node name for the quaternion c parameter. */
  private static String QUATERN_C_NAME = "Quaternion c Parameter";

  /** Output Node description for the quaternion c parameter. */
  private static String QUATERN_C_DESC =
  "C parameter used in the unit quaternion (a,b,c,d) to represent a rotation.";

  /** Output Node name for the quaternion d parameter. */
  private static String QUATERN_D_NAME = "Quaternion d Parameter";

  /** Output Node description for the quaternion d parameter. */
  private static String QUATERN_D_DESC =
  "D parameter used in the unit quaternion (a,b,c,d) to represent a rotation.";

  /** Output Node name for the X offset added after rotation. */
  private static String QOFFSET_X_NAME = "Q Offset X";

  /** Output Node description for the X offset added after rotation. */
  private static String QOFFSET_X_DESC =
  "Offset added to the X dimension after applying the quaternion rotation.";

  /** Output Node name for the Y offset added after rotation. */
  private static String QOFFSET_Y_NAME = "Q Offset Y";

  /** Output Node description for the Y offset added after rotation. */
  private static String QOFFSET_Y_DESC =
  "Offset added to the Y dimension after applying the quaternion rotation.";

  /** Output Node name for the Z offset added after rotation. */
  private static String QOFFSET_Z_NAME = "Q Offset Z";

  /** Output Node description for the Z offset added after rotation. */
  private static String QOFFSET_Z_DESC =
  "Offset added to the Z dimension after applying the quaternion rotation.";

  /**
   * Constructs a DICOM to NIFTI Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToNiftiSpace(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the row direction cosine
      Class type = Double.class;
      _addInputNode( new InputNode(X_ROW_NAME, X_ROW_DESC, type) );
      _addInputNode( new InputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
      _addInputNode( new InputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

      // Create the Input Nodes for the column direction cosine
      type = Double.class;
      _addInputNode( new InputNode(X_COL_NAME, X_COL_DESC, type) );
      _addInputNode( new InputNode(Y_COL_NAME, Y_COL_DESC, type) );
      _addInputNode( new InputNode(Z_COL_NAME, Z_COL_DESC, type) );

      // Create the Input Nodes for the stack direction cosine
      type = Double.class;
      _addInputNode( new InputNode(X_STK_NAME, X_STK_DESC, type) );
      _addInputNode( new InputNode(Y_STK_NAME, Y_STK_DESC, type) );
      _addInputNode( new InputNode(Z_STK_NAME, Z_STK_DESC, type) );

      // Create the Input Nodes for the image position
      type = Double.class;
      _addInputNode( new InputNode(X_IMG_NAME, X_IMG_DESC, type) );
      _addInputNode( new InputNode(Y_IMG_NAME, Y_IMG_DESC, type) );
      _addInputNode( new InputNode(Z_IMG_NAME, Z_IMG_DESC, type) );

      // Create the Output Nodes for the Q code and scaling
      type = Short.class;
      _addOutputNode(new OutputNode(QFORM_CODE_NAME, QFORM_CODE_DESC, type) );
      type = Float.class;
      _addOutputNode(new OutputNode(QFAC_NAME, QFAC_DESC, type) );

      // Create the Output Nodes for the quaternion parameters
      type = Float.class;
      _addOutputNode(new OutputNode(QUATERN_B_NAME, QUATERN_B_DESC, type) );
      _addOutputNode(new OutputNode(QUATERN_C_NAME, QUATERN_C_DESC, type) );
      _addOutputNode(new OutputNode(QUATERN_D_NAME, QUATERN_D_DESC, type) );

      // Create the Output Nodes for the Q offsets
      type = Float.class;
      _addOutputNode(new OutputNode(QOFFSET_X_NAME, QOFFSET_X_DESC, type) );
      _addOutputNode(new OutputNode(QOFFSET_Y_NAME, QOFFSET_Y_DESC, type) );
      _addOutputNode(new OutputNode(QOFFSET_Z_NAME, QOFFSET_Z_DESC, type) );
    }

  /**
   * Constructs a DICOM to NIFTI Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToNiftiSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("DICOM STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's left, the y axis\n");
      b.append("points to the patient's posterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)     ^ Z\n");
      b.append("     _|_      |\n");
      b.append("      |       | Y\n");
      b.append("     / \\      x -- > X\n");
      b.append("\n");
      b.append("The image plane is defined by two vectors.  Each vector\n");
      b.append("starts at the top left corner of the image plane.  The\n");
      b.append("first vector points along the top row of the image plane.\n");
      b.append("The second vector points along the left column of the\n");
      b.append("image plane:\n");
      b.append("\n");
      b.append("    *-----------------> 1st vector\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&& Image Plane &&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    V \n");
      b.append("      2nd vector\n");
      b.append("\n");
      b.append("The row and column vectors are given in the x-y-z\n");
      b.append("coordinate system defined above.  The magnitude of each\n");
      b.append("projection (x,y,z) is given as a cosine of the angle it\n");
      b.append("makes with that projection axis (== direction cosines);\n");
      b.append("e.g:\n");
      b.append("\n");
      b.append(" vector along x axis:\n");
      b.append("  -  0 degree angle with x axis\n");
      b.append("  - 90 degree angle with y axis\n");
      b.append("  - 90 degree angle with z axis\n");
      b.append("\n");
      b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");

      b.append("\n");
      b.append("\n");
      b.append("NIFTI STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's right, the y axis\n");
      b.append("points to the patient's anterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");
      b.append("\n");
      b.append("When quaternions are used to define the coordinate system\n");
      b.append("(qform_code > 0), the formula for computing the\n");
      b.append("coordinate (x,y,z) from the pixel at indices (i,j,k) is:\n");
      b.append("\n");
      b.append("[x]   [R11 R12 R13] [       pixdim[1] * i]   [qoffset_x]\n");
      b.append("[y] = [R21 R22 R23] [       pixdim[2] * j] + [qoffset_y]\n");
      b.append("[z]   [R31 R32 R33] [qfac * pixdim[3] * k]   [qoffset_z]\n");
      b.append("\n");
      b.append("where the element Rij of the rotation matrix are defined\n");
      b.append("using the quaternion parameters (b,c,d).");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the DICOM row directional cosines
	Double rowX = (Double)_getInputValue(X_ROW_NAME, null);
	Double rowY = (Double)_getInputValue(Y_ROW_NAME, null);
	Double rowZ = (Double)_getInputValue(Z_ROW_NAME, null);

	if (rowX == null || rowY == null || rowZ == null) { return; }

	// Get the DICOM column directional cosines
	Double colX = (Double)_getInputValue(X_COL_NAME, null);
	Double colY = (Double)_getInputValue(Y_COL_NAME, null);
	Double colZ = (Double)_getInputValue(Z_COL_NAME, null);

	if (colX == null || colY == null || colZ == null) { return; }

	// Get the DICOM stack directional cosines
	Double stkX = (Double)_getInputValue(X_STK_NAME, null);
	Double stkY = (Double)_getInputValue(Y_STK_NAME, null);
	Double stkZ = (Double)_getInputValue(Z_STK_NAME, null);

	if (stkX == null || stkY == null || stkZ == null) { return; }

	// Get the DICOM image position (use zero if none is supplied)
	Double imgX = (Double)_getInputValue(X_IMG_NAME, new Double(0));
	Double imgY = (Double)_getInputValue(Y_IMG_NAME, new Double(0));
	Double imgZ = (Double)_getInputValue(Z_IMG_NAME, new Double(0));

	// Get the elements of the rotation matrix (negate x,y components)
	double r[][] = new double[3][3];
	r[0][0] = -rowX.doubleValue();
	r[1][0] = -rowY.doubleValue();
	r[2][0] = rowZ.doubleValue();

	r[0][1] = -colX.doubleValue();
	r[1][1] = -colY.doubleValue();
	r[2][1] = colZ.doubleValue();

	r[0][2] = -stkX.doubleValue();
	r[1][2] = -stkY.doubleValue();
	r[2][2] = stkZ.doubleValue();

	// Get the closest normal and orthogonal matrix
	r = _getOrthogonalMatrix(r);

	double r11 = r[0][0];
	double r12 = r[0][1];
	double r13 = r[0][2];

	double r21 = r[1][0];
	double r22 = r[1][1];
	double r23 = r[1][2];

	double r31 = r[2][0];
	double r32 = r[2][1];
	double r33 = r[2][2];

	// Set the Q factor from the determinant (should be -1 or 1)
	double det = _getDeterminant(r);
	if (det > 0) {
	  _setOutputValue(QFAC_NAME, new Float(1));
	}

	// Improper (flip 3rd column)
	else {
	  _setOutputValue(QFAC_NAME, new Float(-1));
	  r13 *= -1;
	  r23 *= -1;
	  r33 *= -1;
	}

	// Compute the quaternion parameters
	double a = r11 + r22 + r33 + 1;
	double b, c, d;

	// 4*a*a not near zero
	if (a > 0.5) {
	  a = 0.5 * Math.sqrt(a);
	  b = 0.25*(r32-r23)/a;
	  c = 0.25*(r13-r31)/a;
	  d = 0.25*(r21-r12)/a;
	}

	// 4*a*a near zero
	else {
	  double xd = 1 + r11 - (r22+r33);
	  double yd = 1 + r22 - (r11+r33);
	  double zd = 1 + r33 - (r11+r22);

	  // 4*b*b not near zero
	  if (xd > 1) {
	    b = 0.5 * Math.sqrt(xd);
	    c = 0.25*(r12+r21)/b;
	    d = 0.25*(r13+r31)/b;
	    a = 0.25*(r32-r23)/b;
	  }

	  // 4*c*c not near zero
	  else if (yd > 1) {
	    c = 0.5 * Math.sqrt(yd);
	    b = 0.25*(r12+r21)/c;
	    d = 0.25*(r23+r32)/c;
	    a = 0.25*(r13-r31)/c;
	  }

	  // 4*d*d not near zero
	  else {
	    d = 0.5 * Math.sqrt(zd);
	    b = 0.25*(r13+r31)/d;
	    c = 0.25*(r23+r32)/d;
	    a = 0.25*(r21-r12)/d;
	  }

	  if (a < 0) {
	    a *= -1;
	    b *= -1;
	    c *= -1;
	    d *= -1;
	  }
	}

	// Set the b,c,d quaternion parameters
	_setOutputValue(QUATERN_B_NAME, new Float(b));
	_setOutputValue(QUATERN_C_NAME, new Float(c));
	_setOutputValue(QUATERN_D_NAME, new Float(d));

	// Set the Q Form Code to NIFTI_XFORM_SCANNER_ANAT
	_setOutputValue(QFORM_CODE_NAME, new Short((short)1));

	// Set the Q offsets (negate the x and y image positions)
	float qoffsetX = -imgX.floatValue();
	float qoffsetY = -imgY.floatValue();
	float qoffsetZ = imgZ.floatValue();
	_setOutputValue(QOFFSET_X_NAME, new Float(qoffsetX));
	_setOutputValue(QOFFSET_Y_NAME, new Float(qoffsetY));
	_setOutputValue(QOFFSET_Z_NAME, new Float(qoffsetZ));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the normal and orthogonal matrix closest to the specified matrix.
   * This method is from nifti1_io.c which in turn used the algorithm from
   * NJ Higham, SIAM J Sci Stat Comput, 7:1160-1174.
   *
   * @param r Input 3x3 matrix.
   *
   * @return Normal and orthogonal matrix closest to the specified matrix.
   *
   * @throws IllegalArgumentException If a matrix column has zero length or if
   *                                  a matrix inversion error occurs.
   */
  private double[][] _getOrthogonalMatrix(double[][] r)
    {
      double r11 = r[0][0];
      double r12 = r[0][1];
      double r13 = r[0][2];

      double r21 = r[1][0];
      double r22 = r[1][1];
      double r23 = r[1][2];

      double r31 = r[2][0];
      double r32 = r[2][1];
      double r33 = r[2][2];

      // Compute the column lengths
      double xd = Math.sqrt(r11*r11 + r21*r21 + r31*r31);
      double yd = Math.sqrt(r12*r12 + r22*r22 + r32*r32);
      double zd = Math.sqrt(r13*r13 + r23*r23 + r33*r33);

      // Check for zero lengths
      if (xd == 0 || yd == 0 || zd == 0) {
	String msg = "Matrix column has zero length.";
	throw new IllegalArgumentException(msg);
      }

      // Construct a matrix with normalized columns
      double[][] q = new double[3][3];
      q[0][0] = r11/xd;
      q[0][1] = r12/yd;
      q[0][2] = r13/zd;

      q[1][0] = r21/xd;
      q[1][1] = r22/yd;
      q[1][2] = r23/zd;

      q[2][0] = r31/xd;
      q[2][1] = r32/yd;
      q[2][2] = r33/zd;

      // Force the matrix to be non-singular
      double gam = _getDeterminant(q);
      while (gam == 0) {

	// Perturb the matrix
	gam = 0.00001 * (0.001 + _getMaxRowNorm(q)) ;

	// Update the diagonal elements
	q[0][0] += gam;
	q[1][1] += gam;
	q[2][2] += gam;

	// Get the new determinant
	gam = _getDeterminant(q);
      }

      // Get the closest orthogonal matrix
      double gmi;
      double dif = 1;
      int k = 0;

      // Convergence or exhaustion
      while (k < 100 && dif > 0.000001) { 
	double[][] qi = _getInverse(q);

	// Far from convergence
	if (dif > 0.3) {
	  double alp = Math.sqrt(_getMaxRowNorm(q) * _getMaxColumnNorm(q));
	  double bet = Math.sqrt(_getMaxRowNorm(qi) * _getMaxColumnNorm(qi));
	  gam = Math.sqrt(bet/alp);
	  gmi = 1.0/gam;
	}

	// Close to convergence
	else {
	  gam = 1.0;
	  gmi = 1.0;
	}

	// Compute the new matrix
	double[][] z = new double[3][3];

	z[0][0] = 0.5*(gam*q[0][0] + gmi*qi[0][0]);
	z[0][1] = 0.5*(gam*q[0][1] + gmi*qi[1][0]);
	z[0][2] = 0.5*(gam*q[0][2] + gmi*qi[2][0]);

	z[1][0] = 0.5*(gam*q[1][0] + gmi*qi[0][1]);
	z[1][1] = 0.5*(gam*q[1][1] + gmi*qi[1][1]);
	z[1][2] = 0.5*(gam*q[1][2] + gmi*qi[2][1]);

	z[2][0] = 0.5*(gam*q[2][0] + gmi*qi[0][2]);
	z[2][1] = 0.5*(gam*q[2][1] + gmi*qi[1][2]);
	z[2][2] = 0.5*(gam*q[2][2] + gmi*qi[2][2]);

	dif = Math.abs(z[0][0] - q[0][0]) + Math.abs(z[0][1] - q[0][1]) +
	      Math.abs(z[0][2] - q[0][2]) + Math.abs(z[1][0] - q[1][0]) +
	      Math.abs(z[1][1] - q[1][1]) + Math.abs(z[1][2] - q[1][2]) +
	      Math.abs(z[2][0] - q[2][0]) + Math.abs(z[2][1] - q[2][1]) +
	      Math.abs(z[2][2] - q[2][2]);

	// Next iteration
	k++;
	q = z;
      }

      return q;
    }

  /**
   * Gets the inverse of the specified matrix.  This method is from
   * nifti1_io.c.
   *
   * @param r Input 3x3 matrix.
   *
   * @return Inverse of the specified matrix.
   *
   * @throws IllegalArgumentException If the matrix cannot be inverted.
   */
  private double[][] _getInverse(double[][] r)
    {
      double r11 = r[0][0];
      double r12 = r[0][1];
      double r13 = r[0][2];

      double r21 = r[1][0];
      double r22 = r[1][1];
      double r23 = r[1][2];

      double r31 = r[2][0];
      double r32 = r[2][1];
      double r33 = r[2][2];

      // Compute the determinant
      double det = _getDeterminant(r);
      if (det == 0) {
	String msg = "Matrix has a zero determinant.";
	throw new IllegalArgumentException(msg);
      }

      // Compute the inverse
      double[][] q = new double[3][3];

      q[0][0] = (r22*r33 - r32*r23)/det;
      q[0][1] = (-r12*r33 + r32*r13)/det;
      q[0][2] = (r12*r23 - r22*r13)/det;

      q[1][0] = (-r21*r33 + r31*r23)/det;
      q[1][1] = (r11*r33 - r31*r13)/det;
      q[1][2] = (-r11*r23 + r21*r13)/det;

      q[2][0] = (r21*r32 - r31*r22)/det;
      q[2][1] = (-r11*r32 + r31*r12)/det;
      q[2][2] = (r11*r22 - r21*r12)/det;

      // Return the inverse
      return q;
    }

  /**
   * Gets the maximum row norm of the specified matrix.  This method is from
   * nifti1_io.c.
   *
   * @param r Input 3x3 matrix.
   *
   * @return Maximum row norm of the specified matrix.
   */
  private double _getMaxRowNorm(double[][] r)
    {
      double r11 = r[0][0];
      double r12 = r[0][1];
      double r13 = r[0][2];

      double r21 = r[1][0];
      double r22 = r[1][1];
      double r23 = r[1][2];

      double r31 = r[2][0];
      double r32 = r[2][1];
      double r33 = r[2][2];

      double n1 = Math.abs(r11) + Math.abs(r12) + Math.abs(r13);
      double n2 = Math.abs(r21) + Math.abs(r22) + Math.abs(r23);
      double n3 = Math.abs(r31) + Math.abs(r32) + Math.abs(r33);

      // Find and return the maximum
      if (n1 < n2) { n1 = n2; }
      if (n1 < n3) { n1 = n3; }
      return n1;
    }

  /**
   * Gets the maximum column norm of the specified matrix.  This method is from
   * nifti1_io.c.
   *
   * @param r Input 3x3 matrix.
   *
   * @return Maximum column norm of the specified matrix.
   */
  private double _getMaxColumnNorm(double[][] r)
    {
      double r11 = r[0][0];
      double r12 = r[0][1];
      double r13 = r[0][2];

      double r21 = r[1][0];
      double r22 = r[1][1];
      double r23 = r[1][2];

      double r31 = r[2][0];
      double r32 = r[2][1];
      double r33 = r[2][2];

      double n1 = Math.abs(r11) + Math.abs(r21) + Math.abs(r31);
      double n2 = Math.abs(r12) + Math.abs(r22) + Math.abs(r32);
      double n3 = Math.abs(r13) + Math.abs(r23) + Math.abs(r33);

      // Find and return the maximum
      if (n1 < n2) { n1 = n2; }
      if (n1 < n3) { n1 = n3; }
      return n1;
    }

  /**
   * Gets the determinant of the specified matrix.  This method is from
   * nifti1_io.c.
   *
   * @param r Input 3x3 matrix.
   *
   * @return Determinant of the specificed matrix.
   */
  private double _getDeterminant(double[][] r)
    {
      double r11 = r[0][0];
      double r12 = r[0][1];
      double r13 = r[0][2];

      double r21 = r[1][0];
      double r22 = r[1][1];
      double r23 = r[1][2];

      double r31 = r[2][0];
      double r32 = r[2][1];
      double r33 = r[2][2];

      return r11*r22*r33 - r11*r32*r23 - r21*r12*r33 +
	     r21*r32*r13 + r31*r12*r23 - r31*r22*r13;
    }
}
