/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.TreeInputNode;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Processor that inserts values into a DICOM tree.
 *
 * @version 6 April 2005
 */
public class DicomTreeValueInsertor extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Inserts values into a DICOM data tree.";

  /** Input Node name for the DICOM data tree. */
  private static String INPUT_TREE_NAME = "DICOM Tree";

  /** Input Node description for the DICOM data tree. */
  private static String INPUT_TREE_DESC = "DICOM Tree to insert values into.";

  /** Input Node name for the value array. */
  private static String VALUES_NAME = "Value Array";

  /** Input Node description for the value array. */
  private static String VALUES_DESC =
  "Array of values to insert into the tree.";

  /** Output Node name for the DICOM data tree with inserted values. */
  private static String OUTPUT_TREE_NAME = "DICOM Tree";

  /** Output Node description for the DICOM data tree with inserted values. */
  private static String OUTPUT_TREE_DESC = "DICOM Tree with inserted values.";

  /** Index of the value array. */
  private int _valueIndex;

  /**
   * Constructs a DICOM Tree Value Insertor with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeValueInsertor(String name)
    {
      super(name, DESC);

      // Create the Input Node for the DICOM data tree
      Class type = MetadataTree.class;
      _addInputNode( new InputNode(INPUT_TREE_NAME, INPUT_TREE_DESC, type) );

      // Create the Input Node for the values
      type = Object[].class;
      _addInputNode( new InputNode(VALUES_NAME, VALUES_DESC, type) );

      // Create the Output Node
      type = MetadataTree.class;
      _addOutputNode(new OutputNode(OUTPUT_TREE_NAME, OUTPUT_TREE_DESC, type));
    }

  /**
   * Constructs a DICOM Tree Value Insertor using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeValueInsertor(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the DICOM tree
	MetadataTree dcmTree = (MetadataTree)_getInputValue(INPUT_TREE_NAME,
							    null);
	if (dcmTree == null) { return; }

	// Clone the DICOM tree
	MetadataTree cloneTree = (MetadataTree)dcmTree.clone();
	cloneTree.setTreeInputNodeValues(dcmTree);

	// Set the values
	Object[] values = (Object[])_getInputValue(VALUES_NAME, null);
	if (values != null) {
	  _valueIndex = 0;
	  _setValues(cloneTree, values);
	}

	// Set the output tree
	_setOutputValue(OUTPUT_TREE_NAME, cloneTree);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Sets the values in the specified Metadata Tree using values in the given
   * array.
   *
   * @param tree Tree to insert values into.
   * @param values Values to insert.
   */
  private void _setValues(MetadataTree tree, Object[] values)
    {
      // Look for a value
      TreeInputNode inputNode = tree.getPrimaryInputNode("value");
      if (inputNode != null && _valueIndex < values.length) {
	inputNode.setValue(values[_valueIndex]);
	_valueIndex++;
      }
	
      // Otherwise search the children for values
      else {
	Enumeration enum = tree.getModules();
	while ( enum.hasMoreElements() ) {
	  Object childTree = enum.nextElement();

	  // Metadata Tree
	  if (childTree instanceof MetadataTree) {
	    _setValues((MetadataTree)childTree, values);
	  }

	  // Metadata Tree Array
	  else if (childTree instanceof MetadataTreeArray) {
	    MetadataTreeArray array = (MetadataTreeArray)childTree;
	    for (int i = 0; i < array.getArraySize(); i++) {
	      _setValues(array.getMetadataTree(i), values);
	    }
	  }
	}
      }
    }
}
