/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that converts from MINC space to DICOM space.
 *
 * @version 30 July 2003
 */
public class MincToDicomSpace extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Converts from MINC space to DICOM space.";

  /** Input Node name for the name of the fastest varying dimension V1. */
  private static String V1_NAME = "Row Name";

  /** Input Node description for the name of the fastest varying dimension. */
  private static String V1_DESC = "MINC name for the fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 2nd fastest varying dimension V2. */
  private static String V2_NAME = "Column Name";

  /** Input Node description for the name of the 2nd fastest dimension. */
  private static String V2_DESC = "MINC name for the 2nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 3nd fastest varying dimension V3. */
  private static String V3_NAME = "Stack Name";

  /** Input Node description for the name of the 3nd fastest dimension. */
  private static String V3_DESC = "MINC name for the 3nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the MINC xspace start location. */
  private static String XSPACE_START_NAME = "Xspace Start Location";

  /** Input Node description for the MINC xspace start location. */
  private static String XSPACE_START_DESC = "Starting location along the " +
                                            "xspace direction.";

  /** Input Node name for the MINC xspace step size. */
  private static String XSPACE_STEP_NAME = "Xspace Step";

  /** Input Node description for the MINC xspace step size. */
  private static String XSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the xspace direction.";

  /** Input Node name for the MINC xspace X direction cosine. */
  private static String XSPACE_X_NAME = "Xspace X Direction Cosine";

  /** Input Node description for the MINC xspace X direction cosine. */
  private static String XSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_NAME = "Xspace Y Direction Cosine";

  /** Input Node description for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_NAME = "Xspace Z Direction Cosine";

  /** Input Node description for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC yspace start location. */
  private static String YSPACE_START_NAME = "Yspace Start Location";

  /** Input Node description for the MINC yspace start location. */
  private static String YSPACE_START_DESC = "Starting location along the " +
                                            "yspace direction.";

  /** Input Node name for the MINC yspace step size. */
  private static String YSPACE_STEP_NAME = "Yspace Step";

  /** Input Node description for the MINC yspace step size. */
  private static String YSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the yspace direction.";

  /** Input Node name for the MINC yspace X direction cosine. */
  private static String YSPACE_X_NAME = "Yspace X Direction Cosine";

  /** Input Node description for the MINC yspace X direction cosine. */
  private static String YSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_NAME = "Yspace Y Direction Cosine";

  /** Input Node description for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_NAME = "Yspace Z Direction Cosine";

  /** Input Node description for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC zspace start location. */
  private static String ZSPACE_START_NAME = "Zspace Start Location";

  /** Input Node description for the MINC zspace start location. */
  private static String ZSPACE_START_DESC = "Starting location along the " +
                                            "zspace direction.";

  /** Input Node name for the MINC zspace step size. */
  private static String ZSPACE_STEP_NAME = "Zspace Step";

  /** Input Node description for the MINC zspace step size. */
  private static String ZSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the zspace direction.";

  /** Input Node name for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_NAME = "Zspace X Direction Cosine";

  /** Input Node description for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_NAME = "Zspace Y Direction Cosine";

  /** Input Node description for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_NAME = "Zspace Z Direction Cosine";

  /** Input Node description for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_DESC = "Direction cosine along Z.";

  /** Output Node name for the X coordinate of the row direction cosine. */
  private static String X_ROW_NAME = "Row Direction Cosine (X)";

  /** Output Node description for the X coord of the row direction cosine. */
  private static String X_ROW_DESC = "X coord of the row direction cosine.";

  /** Output Node name for the Y coordinate of the row direction cosine. */
  private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

  /** Output Node description for the Y coord of the row direction cosine. */
  private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

  /** Output Node name for the Z coordinate of the row direction cosine. */
  private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

  /** Output Node description for the Z coord of the row direction cosine. */
  private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

  /** Output Node name for the X coordinate of the column direction cosine. */
  private static String X_COL_NAME = "Column Direction Cosine (X)";

  /** Output Node description for X coord of the column direction cosine. */
  private static String X_COL_DESC = "X coord of the column direction cosine.";

  /** Output Node name for the Y coordinate of the column direction cosine. */
  private static String Y_COL_NAME = "Column Direction Cosine (Y)";

  /** Output Node description for Y coord of the column direction cosine. */
  private static String Y_COL_DESC = "Y coord of the column direction cosine.";

  /** Output Node name for the Z coordinate of the column direction cosine. */
  private static String Z_COL_NAME = "Column Direction Cosine (Z)";

  /** Output Node description for Z coord of the column direction cosine. */
  private static String Z_COL_DESC = "Z coord of the column direction cosine.";

  /** Output Node name for the X coordinate of the stack direction cosine. */
  private static String X_STK_NAME = "Stack Direction Cosine (X)";

  /** Output Node description for X coord of the stack direction cosine. */
  private static String X_STK_DESC = "X coord of the stack direction cosine.";

  /** Output Node name for the Y coordinate of the stack direction cosine. */
  private static String Y_STK_NAME = "Stack Direction Cosine (Y)";

  /** Output Node description for Y coord of the stack direction cosine. */
  private static String Y_STK_DESC = "Y coord of the stack direction cosine.";

  /** Output Node name for the Z coordinate of the stack direction cosine. */
  private static String Z_STK_NAME = "Stack Direction Cosine (Z)";

  /** Output Node description for Z coord of the stack direction cosine. */
  private static String Z_STK_DESC = "Z coord of the stack direction cosine.";

  /** Output Node name for the row spacing. */
  private static String ROW_STEP_NAME = "Row Spacing";

  /** Output Node description for the row spacing. */
  private static String ROW_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the row direction.";

  /** Output Node name for the column spacing. */
  private static String COL_STEP_NAME = "Column Spacing";

  /** Output Node description for the column spacing. */
  private static String COL_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the column direction.";

  /** Output Node name for the stack spacing. */
  private static String STK_STEP_NAME = "Stack Spacing";

  /** Output Node description for the stack spacing. */
  private static String STK_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the stack direction.";

  /** Output Node name for the row start location. */
  private static String ROW_START_NAME = "Row Start Location";

  /** Output Node description for the row start location. */
  private static String ROW_START_DESC = "Starting location along the row " +
                                         "direction.";

  /** Output Node name for the column start location. */
  private static String COL_START_NAME = "Column Start Location";

  /** Output Node description for the column start location. */
  private static String COL_START_DESC = "Starting location along the column" +
                                         " direction.";

  /** Output Node name for the stack start location. */
  private static String STK_START_NAME = "Stack Start Location";

  /** Output Node description for the stack start location. */
  private static String STK_START_DESC = "Starting location along the stack " +
                                         "direction.";

  /**
   * Constructs a MINC to DICOM Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MincToDicomSpace(String name)
    {
      super(name, DESC);

      // Create the "xspace, yspace, zspace" only Restriction
      name = "MINC Dimension Restriction";
      String desc = "Allowed values are xspace, yspace, zspace.";
      Vector options = new Vector(3);
      options.addElement("xspace");
      options.addElement("yspace");
      options.addElement("zspace");
      SetRestriction dimRestrict = new SetRestriction(name, desc, options);

      // Create the Input Nodes for the names of the dimensions
      Class type = String.class;
      InputNode inputNode = new InputNode(V1_NAME, V1_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V2_NAME, V2_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V3_NAME, V3_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      // Create the Input Nodes for the xspace quantities
      type = Double.class;
      _addInputNode(new InputNode(XSPACE_START_NAME, XSPACE_START_DESC, type));
      _addInputNode(new InputNode(XSPACE_STEP_NAME, XSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(XSPACE_X_NAME, XSPACE_X_DESC, type));
      _addInputNode(new InputNode(XSPACE_Y_NAME, XSPACE_Y_DESC, type));
      _addInputNode(new InputNode(XSPACE_Z_NAME, XSPACE_Z_DESC, type));

      // Create the Input Nodes for the yspace quantities
      type = Double.class;
      _addInputNode(new InputNode(YSPACE_START_NAME, YSPACE_START_DESC, type));
      _addInputNode(new InputNode(YSPACE_STEP_NAME, YSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(YSPACE_X_NAME, YSPACE_X_DESC, type));
      _addInputNode(new InputNode(YSPACE_Y_NAME, YSPACE_Y_DESC, type));
      _addInputNode(new InputNode(YSPACE_Z_NAME, YSPACE_Z_DESC, type));

      // Create the Input Nodes for the zspace quantities
      type = Double.class;
      _addInputNode(new InputNode(ZSPACE_START_NAME, ZSPACE_START_DESC, type));
      _addInputNode(new InputNode(ZSPACE_STEP_NAME, ZSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(ZSPACE_X_NAME, ZSPACE_X_DESC, type));
      _addInputNode(new InputNode(ZSPACE_Y_NAME, ZSPACE_Y_DESC, type));
      _addInputNode(new InputNode(ZSPACE_Z_NAME, ZSPACE_Z_DESC, type));

      // Create the Output Nodes for the row quantities
      type = Double.class;
      _addOutputNode( new OutputNode(ROW_START_NAME, ROW_START_DESC, type) );
      _addOutputNode( new OutputNode(ROW_STEP_NAME, ROW_STEP_DESC, type) );

      // Create the Output Nodes for the row direction cosine
      type = Double.class;
      _addOutputNode( new OutputNode(X_ROW_NAME, X_ROW_DESC, type) );
      _addOutputNode( new OutputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
      _addOutputNode( new OutputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

      // Create the Output Nodes for the column quantities
      type = Double.class;
      _addOutputNode( new OutputNode(COL_START_NAME, COL_START_DESC, type) );
      _addOutputNode( new OutputNode(COL_STEP_NAME, COL_STEP_DESC, type) );

      // Create the Output Nodes for the column direction cosine
      type = Double.class;
      _addOutputNode( new OutputNode(X_COL_NAME, X_COL_DESC, type) );
      _addOutputNode( new OutputNode(Y_COL_NAME, Y_COL_DESC, type) );
      _addOutputNode( new OutputNode(Z_COL_NAME, Z_COL_DESC, type) );

      // Create the Output Nodes for the stack quantities
      type = Double.class;
      _addOutputNode( new OutputNode(STK_START_NAME, STK_START_DESC, type) );
      _addOutputNode( new OutputNode(STK_STEP_NAME, STK_STEP_DESC, type) );

      // Create the Output Nodes for the stack direction cosine
      type = Double.class;
      _addOutputNode( new OutputNode(X_STK_NAME, X_STK_DESC, type) );
      _addOutputNode( new OutputNode(Y_STK_NAME, Y_STK_DESC, type) );
      _addOutputNode( new OutputNode(Z_STK_NAME, Z_STK_DESC, type) );
    }

  /**
   * Constructs a MINC to DICOM Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MincToDicomSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("MINC STANDARDS:\n");
      b.append("\n");
      b.append("Voxel coordinates are defined as follows:\n");
      b.append("\n");
      b.append("V1 = Fastest varying dimension\n");
      b.append("V2 = Second fastest varying dimension\n");
      b.append("V3 = Third fastest varying dimension\n");
      b.append("\n");
      b.append("            ^\n");
      b.append("            V3                                  ^\n");
      b.append("            |                     *------------ V1\n");
      b.append("            |                     |\n");
      b.append("  image #2  |                     | ###########  rows ->\n");
      b.append("  image #1  |                     | ###########\n");
      b.append("  image #0  |           ^         | ###########  columns\n");
      b.append("            *---------- V1        |                 |\n");
      b.append("             \\                  ^                   V\n");
      b.append("              \\                 V2\n");
      b.append("               \\ ^\n");
      b.append("                 V2\n");
      b.append("\n");
      b.append("Each coordinate is given a name (\"xspace\", \"yspace\",\n");
      b.append("or \"zspace\") to indicate the patient's orientation.\n");
      b.append("Each name by itself does not indicate the orientation; it\n");
      b.append("is merely a label that denotes a set of directional\n");
      b.append("cosines, step size, and starting location (origin).  It is\n");
      b.append("then possible, for example, for \"xspace\" to denote the\n");
      b.append("Y direction if its directional cosines point in that\n");
      b.append("direction.  Note that if the step size is negative, then\n");
      b.append("the direction is reversed.\n");
      b.append("\n");
      b.append("MINC puts its coordinates in Talairach space so that\n");
      b.append("the x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");

      b.append("\n");
      b.append("\n");
      b.append("DICOM STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's left, the y axis\n");
      b.append("points to the patient's posterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)     ^ Z\n");
      b.append("     _|_      |\n");
      b.append("      |       | Y\n");
      b.append("     / \\      x -- > X\n");
      b.append("\n");
      b.append("The image plane is defined by two vectors.  Each vector\n");
      b.append("starts at the top left corner of the image plane.  The\n");
      b.append("first vector points along the top row of the image plane.\n");
      b.append("The second vector points along the left column of the\n");
      b.append("image plane:\n");
      b.append("\n");
      b.append("    *-----------------> 1st vector\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&& Image Plane &&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    V \n");
      b.append("      2nd vector\n");
      b.append("\n");
      b.append("The row and column vectors are given in the x-y-z\n");
      b.append("coordinate system defined above.  The magnitude of each\n");
      b.append("projection (x,y,z) is given as a cosine of the angle it\n");
      b.append("makes with that projection axis (== direction cosines);\n");
      b.append("e.g:\n");
      b.append("\n");
      b.append(" vector along x axis:\n");
      b.append("  -  0 degree angle with x axis\n");
      b.append("  - 90 degree angle with y axis\n");
      b.append("  - 90 degree angle with z axis\n");
      b.append("\n");
      b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the V1, V2, V3 dimension names
	String v1Name = (String)_getInputValue(V1_NAME, null);
	String v2Name = (String)_getInputValue(V2_NAME, null);
	String v3Name = (String)_getInputValue(V3_NAME, null);

	// Not enough information
	if (v1Name == null || v2Name == null || v3Name == null) {
	  _setOutputValues(null, null, null, null, null, null, null, null,
			   null);
	  return;
	}

	// Get the xspace quantities
	Double xspaceStart = (Double)_getInputValue(XSPACE_START_NAME, null);
	Double xspaceStep = (Double)_getInputValue(XSPACE_STEP_NAME, null);
	Double xspaceX = (Double)_getInputValue(XSPACE_X_NAME, null);
	Double xspaceY = (Double)_getInputValue(XSPACE_Y_NAME, null);
	Double xspaceZ = (Double)_getInputValue(XSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (xspaceX == null || xspaceY == null || xspaceZ == null) {
	  xspaceX = new Double(1.0);
	  xspaceY = new Double(0.0);
	  xspaceZ = new Double(0.0);
	}

	// Get the yspace quantities
	Double yspaceStart = (Double)_getInputValue(YSPACE_START_NAME, null);
	Double yspaceStep = (Double)_getInputValue(YSPACE_STEP_NAME, null);
	Double yspaceX = (Double)_getInputValue(YSPACE_X_NAME, null);
	Double yspaceY = (Double)_getInputValue(YSPACE_Y_NAME, null);
	Double yspaceZ = (Double)_getInputValue(YSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (yspaceX == null || yspaceY == null || yspaceZ == null) {
	  yspaceX = new Double(0.0);
	  yspaceY = new Double(1.0);
	  yspaceZ = new Double(0.0);
	}

	// Get the zspace quantities
	Double zspaceStart = (Double)_getInputValue(ZSPACE_START_NAME, null);
	Double zspaceStep = (Double)_getInputValue(ZSPACE_STEP_NAME, null);
	Double zspaceX = (Double)_getInputValue(ZSPACE_X_NAME, null);
	Double zspaceY = (Double)_getInputValue(ZSPACE_Y_NAME, null);
	Double zspaceZ = (Double)_getInputValue(ZSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (zspaceX == null || zspaceY == null || zspaceZ == null) {
	  zspaceX = new Double(0.0);
	  zspaceY = new Double(0.0);
	  zspaceZ = new Double(1.0);
	}

	// Assume positive step values if none are provided
	if (xspaceStep == null) { xspaceStep = new Double(1.0); }
	if (yspaceStep == null) { yspaceStep = new Double(1.0); }
	if (zspaceStep == null) { zspaceStep = new Double(1.0); }

	// Zero the starting location if no locations are provided
	if (xspaceStart == null) { xspaceStart = new Double(0.0); }
	if (yspaceStart == null) { yspaceStart = new Double(0.0); }
	if (zspaceStart == null) { zspaceStart = new Double(0.0); }

	// Get the DICOM coordinate information
	Double[] rowDirCosines =
	  _getDicomCoordinate(v1Name, xspaceStep, xspaceX, xspaceY, xspaceZ,
			      yspaceStep, yspaceX, yspaceY, yspaceZ,
			      zspaceStep, zspaceX, zspaceY, zspaceZ);

	Double[] colDirCosines =
	  _getDicomCoordinate(v2Name, xspaceStep, xspaceX, xspaceY, xspaceZ,
			      yspaceStep, yspaceX, yspaceY, yspaceZ,
			      zspaceStep, zspaceX, zspaceY, zspaceZ);

	Double[] stkDirCosines =
	  _getDicomCoordinate(v3Name, xspaceStep, xspaceX, xspaceY, xspaceZ,
			      yspaceStep, yspaceX, yspaceY, yspaceZ,
			      zspaceStep, zspaceX, zspaceY, zspaceZ);

	// Get the DICOM step information
	Double rowStep = _getDicomStep(v1Name, xspaceStep, yspaceStep,
				       zspaceStep);

	Double colStep = _getDicomStep(v2Name, xspaceStep, yspaceStep,
				       zspaceStep);

	Double stkStep = _getDicomStep(v3Name, xspaceStep, yspaceStep,
				       zspaceStep);

	// Get the DICOM start information
	Double rowStart = _getDicomStart(v1Name, xspaceStart, yspaceStart,
					 zspaceStart);

	Double colStart = _getDicomStart(v2Name, xspaceStart, yspaceStart,
					 zspaceStart);

	Double stkStart = _getDicomStart(v3Name, xspaceStart, yspaceStart,
					 zspaceStart);

	// Set the output values
	_setOutputValues(rowDirCosines, rowStart, rowStep,
			 colDirCosines, colStart, colStep,
			 stkDirCosines, stkStart, stkStep);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Sets the output values.
   *
   * @param rowDirCosines Row directional cosines.
   * @param rowStart Start location along the row direction.
   * @param rowStep Spacing along the row direction.
   * @param colDirCosines Column directional cosines.
   * @param colStart Start location along the column direction.
   * @param colStep Spacing along the column direction.
   * @param stkDirCosines Stack directional cosines.
   * @param stkStart Start location along the stack direction.
   * @param stkStep Spacing along the stack direction.
   */
  private void _setOutputValues(Double[] rowDirCosines, Double rowStart,
				Double rowStep,
				Double[] colDirCosines, Double colStart,
				Double colStep,
				Double[] stkDirCosines, Double stkStart,
				Double stkStep)
    {
      // Row quantities
      _setOutputValue(X_ROW_NAME, rowDirCosines[0]);
      _setOutputValue(Y_ROW_NAME, rowDirCosines[1]);
      _setOutputValue(Z_ROW_NAME, rowDirCosines[2]);
      _setOutputValue(ROW_START_NAME, rowStart);
      _setOutputValue(ROW_STEP_NAME, rowStep);

      // Column quantities
      _setOutputValue(X_COL_NAME, colDirCosines[0]);
      _setOutputValue(Y_COL_NAME, colDirCosines[1]);
      _setOutputValue(Z_COL_NAME, colDirCosines[2]);
      _setOutputValue(COL_START_NAME, colStart);
      _setOutputValue(COL_STEP_NAME, colStep);

      // Stack quantities
      _setOutputValue(X_STK_NAME, stkDirCosines[0]);
      _setOutputValue(Y_STK_NAME, stkDirCosines[1]);
      _setOutputValue(Z_STK_NAME, stkDirCosines[2]);
      _setOutputValue(STK_START_NAME, stkStart);
      _setOutputValue(STK_STEP_NAME, stkStep);
    }

  /**
   * Gets the DICOM coordinate from the specified MINC coordinate.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param xspaceX Xspace direction cosine in the X direction.
   * @param xspaceY Xspace direction cosine in the Y direction.
   * @param xspaceZ Xspace direction cosine in the Z direction.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param yspaceX Yspace direction cosine in the X direction.
   * @param yspaceY Yspace direction cosine in the Y direction.
   * @param yspaceZ Yspace direction cosine in the Z direction.
   * @param zspaceStep Zspace step size between adjacent voxels.
   * @param zspaceX Zspace direction cosine in the X direction.
   * @param zspaceY Zspace direction cosine in the Y direction.
   * @param zspaceZ Zspace direction cosine in the Z direction.
   *
   * @return DICOM directional cosines from the specified MINC coordinate.
   */
  private Double[] _getDicomCoordinate(String name, Double xspaceStep,
				       Double xspaceX, Double xspaceY,
				       Double xspaceZ, Double yspaceStep,
				       Double yspaceX, Double yspaceY,
				       Double yspaceZ, Double zspaceStep,
				       Double zspaceX, Double zspaceY,
				       Double zspaceZ)
    {
      double cosX = 0;
      double cosY = 0;
      double cosZ = 0;
      double step = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	cosX = xspaceX.doubleValue();
	cosY = xspaceY.doubleValue();
	cosZ = xspaceZ.doubleValue();
	step = xspaceStep.doubleValue();
      }

      // Yspace
      if ( name.equals("yspace") ) {
	cosX = yspaceX.doubleValue();
	cosY = yspaceY.doubleValue();
	cosZ = yspaceZ.doubleValue();
	step = yspaceStep.doubleValue();
      }

      // Zspace
      if ( name.equals("zspace") ) {
	cosX = zspaceX.doubleValue();
	cosY = zspaceY.doubleValue();
	cosZ = zspaceZ.doubleValue();
	step = zspaceStep.doubleValue();
      }

      // Reverse the direction if the step is negative
      if (step < 0) {
	cosX *= -1;
	cosY *= -1;
	cosZ *= -1;
      }

      // DICOM has x,y coordinates flipped
      cosX *= -1;
      cosY *= -1;

      // Return the DICOM coordinate
      Double[] cosines = new Double[3];
      cosines[0] = new Double(cosX);
      cosines[1] = new Double(cosY);
      cosines[2] = new Double(cosZ);

      return cosines;
    }

  /**
   * Gets the DICOM step size from the specified MINC coordinate.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param zspaceStep Zspace step size between adjacent voxels.
   *
   * @return DICOM step size from the specified MINC coordinate.
   */
  private Double _getDicomStep(String name, Double xspaceStep,
			       Double yspaceStep, Double zspaceStep)
    {
      double step = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	step = xspaceStep.doubleValue();
      }

      // Yspace
      if ( name.equals("yspace") ) {
	step = yspaceStep.doubleValue();
      }

      // Zspace
      if ( name.equals("zspace") ) {
	step = zspaceStep.doubleValue();
      }

      // Reverse the direction if the step is negative
      if (step < 0) { step *= -1; }

      // Return the DICOM step size
      return new Double(step);
    }

  /**
   * Gets the DICOM starting location from the specified MINC coordinate.
   *
   * @param name Name of the dimension.
   * @param xspaceStart Xspace start location between adjacent voxels.
   * @param yspaceStart Yspace start location between adjacent voxels.
   * @param zspaceStart Zspace start location between adjacent voxels.
   *
   * @return DICOM start location from the specified MINC coordinate.
   */
  private Double _getDicomStart(String name, Double xspaceStart,
				Double yspaceStart, Double zspaceStart)
    {
      double start = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	start = xspaceStart.doubleValue();
      }

      // Yspace
      if ( name.equals("yspace") ) {
	start = yspaceStart.doubleValue();
      }

      // Zspace
      if ( name.equals("zspace") ) {
	start = zspaceStart.doubleValue();
      }

      // Return the DICOM start location
      return new Double(start);
    }
}
