/*
  COPYRIGHT NOTICE
  Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.math;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Constructor;
import java.util.Vector;

/**
 * Stateful Processor that holds a sum or product inside a Loop Module.
 *
 * @version 15 August 2005
 */
public class NumberAccumulator extends Processor implements Stateful
{
    /** Description of this Processor. */
    private static String DESC = "Computes a sum/product inside a loop.";

    /** Input Node name for the value to add/multiply. */
    private static String VALUE_NAME = "Value";

    /** Property for the operation to perform. */
    private static String PROP_NAME = "operation";

    /** Property description for the operation to perform. */
    private static String PROP_DESC = "Type of operation to perform.";

    /** Property for the value type. */
    private static String PROP_TYPE_NAME = "type";

    /** Property description for the value type. */
    private static String PROP_TYPE_DESC = "Type of values to accumulate.";

    /** Accumulated integral sum/product. */
    private long _integralStore;

    /** Accumulated decimal sum/product. */
    private double _decimalStore;

    /**
     * Constructs a Number Accumulator with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NumberAccumulator(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a Number Accumulator using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NumberAccumulator(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Get the operation to perform and the data type
	String operation = (String)_getPropertyValue(PROP_NAME);
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);

	// Create the Input Node
	String desc = "Value to multiply to the product.";
	if ( "+".equals(operation) ) { desc = "Value to add to the sum."; }
	_addInputNode( new InputNode(VALUE_NAME, desc, type) );

	// Create the Output Node
	name = "Product";
	desc = "Product of all the inputs.";
	if ( "+".equals(operation) ) {
	    name = "Sum";
	    desc = "Sum of all the inputs.";
	}
	_addOutputNode( new OutputNode(name, desc, type) );
    }

    /**
     * Reinitializes the state.  This is called by a Loop Module before a loop
     * is started.
     */
    public void reinitializeState()
    {
	String operation = (String)_getPropertyValue(PROP_NAME);

	// Initialize the stores according to the operation
	_integralStore = 1;
	_decimalStore = 1;
	if ( "+".equals(operation) ) {
	    _integralStore = 0;
	    _decimalStore = 0;
	}
    }

    /**
     * Gets the long description.
     *
     * @return A verbose description of what the Processor does.
     */
    public String getLongDescription()
    {
	StringBuffer b = new StringBuffer();

	b.append("Note:  When the loop begins, the internal store is set to\n");
	b.append("zero (when adding) or one (when multiplying).");

	return b.toString();
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the operation to perform and the data type
	    String operation = (String)_getPropertyValue(PROP_NAME);
	    Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);

	    // Get the value to add/multiply
	    Object value = _getInputValue(VALUE_NAME, null);
	    if (value != null) {

		// Add the value to the store
		if ( "+".equals(operation) ) {
		    _integralStore += ((Number)value).longValue();
		    _decimalStore += ((Number)value).doubleValue();
		}

		// Or multiply the value to the store
		else {
		    _integralStore *= ((Number)value).longValue();
		    _decimalStore *= ((Number)value).doubleValue();
		}
	    }

	    // Get the Constructor of the type with one String argument
	    Class[] parameters = new Class[1];
	    parameters[0] = String.class;
	    Constructor constructor = type.getDeclaredConstructor(parameters);

	    // Set the appropriate Output Node value
	    Object[] arguments = new Object[1];
	    if (type == Float.class || type == Double.class) {
		arguments[0] = _decimalStore + "";
	    }
	    else { arguments[0] = _integralStore + ""; }

	    // Set the Output Node
	    String name = "Product";
	    if ( "+".equals(operation) ) { name = "Sum"; }
	    Object constant = constructor.newInstance(arguments);
	    _setOutputValue(name, constant);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \""+this+"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[2];

	// Allowed operations
	Vector optionSet = new Vector();
	optionSet.add("+");
	optionSet.add("*");

	// Create the operation restriction
	String name = "Allowed Operations";
	String desc = "Operations allowed for accumulating.";
	SetRestriction restriction = new SetRestriction(name, desc, optionSet);

	// Default operation
	name = PROP_NAME;
	desc = PROP_DESC;
	Object value = "+";
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	// Allowed primitive types
	optionSet = new Vector();
	optionSet.add( Byte.class );
	optionSet.add( Double.class );
	optionSet.add( Float.class );
	optionSet.add( Integer.class );
	optionSet.add( Long.class );
	optionSet.add( Short.class );

	// Create the type restriction
	name = "Allowed Types";
	desc = "Types allowed for accumulating values.";
	restriction = new SetRestriction(name, desc, optionSet);

	// Default value type
	name = PROP_TYPE_NAME;
	desc = PROP_TYPE_DESC;
	value = Double.class;
	properties[1] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
