/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Stateful Processor that sorts DICOM images.
 *
 * @version 15 December 2003
 */
public class DicomImageSorter extends Processor implements Stateful
{
  /** Description of this Processor. */
  private static String DESC = "Sorts DICOM images.";

  /** Input Node name for the X coordinate of the first image pixel. */
  private static String X_NAME = "Image Position (X)";

  /** Input Node description for the X coordinate of the first image pixel. */
  private static String X_DESC = "X coordinate of the first image pixel.";

  /** Input Node name for the Y coordinate of the first image pixel. */
  private static String Y_NAME = "Image Position (Y)";

  /** Input Node description for the Y coordinate of the first image pixel. */
  private static String Y_DESC = "Y coordinate of the first image pixel.";

  /** Input Node name for the Z coordinate of the first image pixel. */
  private static String Z_NAME = "Image Position (Z)";

  /** Input Node description for the Z coordinate of the first image pixel. */
  private static String Z_DESC = "Z coordinate of the first image pixel.";

  /** Input Node name for the Talairach coordinate to sort along. */
  private static String SORT_NAME = "Sorting Talairach Coordinate";

  /** Input Node description for the Talairach coordinate to sort along. */
  private static String SORT_DESC = "Talairach Coordinate to sort along.";

  /** Input Node name for the image. */
  private static String IMAGE_NAME = "Image";
  
  /** Input Node description for the image. */
  private static String IMAGE_DESC = "Image to sort.";

  /** Output Node name for the sorted image array. */
  private static String ARRAY_NAME = "Sorted Image Array.";

  /** Output Node description for the sorted image array. */
  private static String ARRAY_DESC = "Images sorted by DICOM field.";

  /** Output Node name for the average slice spacing. */
  private static String SPACING_NAME = "Average Slice Spacing";

  /** Output Node description for the average slice spacing. */
  private static String SPACING_DESC = "Average spacing between slices.";

  /** Output Node name for the X coordinate of the first sorted image pixel. */
  private static String ORIGIN_X_NAME = "Origin (X)";

  /** Output Node desc for the X coordinate of the first sorted image pixel. */
  private static String ORIGIN_X_DESC =
  "X coordinate of the first sorted image's pixel.";

  /** Output Node name for the Y coordinate of the first sorted image pixel. */
  private static String ORIGIN_Y_NAME = "Origin (Y)";

  /** Output Node desc for the Y coordinate of the first sorted image pixel. */
  private static String ORIGIN_Y_DESC =
  "Y coordinate of the first sorted image's pixel.";

  /** Output Node name for the Z coordinate of the first sorted image pixel. */
  private static String ORIGIN_Z_NAME = "Origin (Z)";

  /** Output Node desc for the Z coordinate of the first sorted image pixel. */
  private static String ORIGIN_Z_DESC =
  "Z coordinate of the first sorted image's pixel.";

  /** Sorted Dicom Images. */
  private Vector _sortedImages;

  /** Talairach coordinate to sort along. */
  private String _sortingCoordinate;

  /**
   * Constructs a Dicom Image Sorter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomImageSorter(String name)
    {
      super(name, DESC);

      // Create the Input Node for the X coordinate of the first image pixel
      Class type = Double.class;
      InputNode inputNode = new InputNode(X_NAME, X_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Y coordinate of the first image pixel
      type = Double.class;
      inputNode = new InputNode(Y_NAME, Y_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Z coordinate of the first image pixel
      type = Double.class;
      inputNode = new InputNode(Z_NAME, Z_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Talairach coordinate to sort along
      type = String.class;
      inputNode = new InputNode(SORT_NAME, SORT_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the image
      type = BufferedImage.class;
      inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the sorted image array
      type = BufferedImage[].class;
      OutputNode node = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the average slice spacing
      type = Double.class;
      node = new OutputNode(SPACING_NAME, SPACING_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the X coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_X_NAME, ORIGIN_X_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the Y coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_Y_NAME, ORIGIN_Y_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the Z coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_Z_NAME, ORIGIN_Z_DESC, type);
      _addOutputNode(node);

      // Clear the sorted images
      reinitializeState();
    }

  /**
   * Constructs a Dicom Image Sorter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomImageSorter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input values
	Double x = (Double)_getInputValue(X_NAME, null);
	Double y = (Double)_getInputValue(Y_NAME, null);
	Double z = (Double)_getInputValue(Z_NAME, null);

	_sortingCoordinate = (String)_getInputValue(SORT_NAME, null);
	if (_sortingCoordinate == null) {
	  String msg = "No Talairach coordinate to sort along.";
	  throw new ProcessingException(msg);
	}

	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);
	if (image == null) {
	  String msg = "No image to sort.";
	  throw new ProcessingException(msg);
	}

	// Construct a Dicom Image from the Input values
	DicomImage dicomImage = new DicomImage(x, y, z, image);
	_sortedImages.add(dicomImage);

	// Sort the Dicom Images
	Collections.sort(_sortedImages, _getComparator());

	// Set the array of sorted Buffered Images
	BufferedImage[] images = new BufferedImage[ _sortedImages.size() ];
	for (int i = 0; i < _sortedImages.size(); i++) {
	  dicomImage = (DicomImage)_sortedImages.elementAt(i);
	  images[i] = dicomImage.getBufferedImage();
	}
	_setOutputValue(ARRAY_NAME, images);

	// Set the average slice spacing
	double sliceSpacing = _computeAverageSliceSpacing(_sortedImages);
	_setOutputValue(SPACING_NAME, new Double(sliceSpacing));

	// Set the origin
	if ( _sortedImages.size() > 0 ) {

	  // Set the coordinates of the first sorted image
	  dicomImage = (DicomImage)_sortedImages.elementAt(0);
	  _setOutputValue(ORIGIN_X_NAME, dicomImage.getX());
	  _setOutputValue(ORIGIN_Y_NAME, dicomImage.getY());
	  _setOutputValue(ORIGIN_Z_NAME, dicomImage.getZ());
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Reinitializes the state.  This is called by a Loop Module before a loop is
   * started.
   */
  public void reinitializeState()
    {
      _sortedImages = new Vector();
    }

  /**
   * Computes the average slice spacing between the ordered DICOM images.
   *
   * @param sortedImages Sorted DICOM images.
   *
   * @return Average slice spacing between the ordered DICOM images.
   */
  private double _computeAverageSliceSpacing(Vector sortedImages)
    {
      double sliceSpacing = 0;

      // Get the average slice spacing
      int numberOfImages = sortedImages.size();
      for (int i = 0; i < numberOfImages-1; i++) {
	DicomImage d1 = (DicomImage)sortedImages.elementAt(i);
	DicomImage d2 = (DicomImage)sortedImages.elementAt(i+1);

	// Determine if the Image Position coordinates exists
	if ( d1.getX() != null && d1.getY() != null &&
	     d1.getZ() != null && d2.getX() != null && 
	     d2.getY() != null && d2.getZ() != null )
	  {
	    // Variation along x direction
	    if ( _sortingCoordinate.equals(Talairach.POSITIVE_X) ||
		 _sortingCoordinate.equals(Talairach.NEGATIVE_X) )
	      {
		double x1 = d1.getX().doubleValue();
		double x2 = d2.getX().doubleValue();
		sliceSpacing += Math.abs(x2-x1);
	      }

	    // Variation along y direction
	    if ( _sortingCoordinate.equals(Talairach.POSITIVE_Y) ||
		 _sortingCoordinate.equals(Talairach.NEGATIVE_Y) )
	      {
		double y1 = d1.getY().doubleValue();
		double y2 = d2.getY().doubleValue();
		sliceSpacing += Math.abs(y2-y1);
	      }

	    // Variation along z direction
	    if ( _sortingCoordinate.equals(Talairach.POSITIVE_Z) ||
		 _sortingCoordinate.equals(Talairach.NEGATIVE_Z) )
	      {
		double z1 = d1.getZ().doubleValue();
		double z2 = d2.getZ().doubleValue();
		sliceSpacing += Math.abs(z2-z1);
	      }
	  }
      }

      // Return the average slice spacing
      int numberOfSpaces = numberOfImages - 1;
      if (numberOfSpaces > 0) { sliceSpacing /= numberOfSpaces; }
      return sliceSpacing;
    }

  /**
   * Gets a Comparator for sorting DICOM images.
   *
   * @return Comparator for sorting DICOM images.
   */
  private Comparator _getComparator()
    {
      return new Comparator()
	{
	  /**
	   * Compares its two arguments for order.
	   *
	   * @param o1 The first object to be compared.
	   * @param o2 The second object to be compared.
	   *
	   * @return A negative integer, zero, or a positive integer as the
	   *         first argument is less than, equal to, or greater than
	   *         the second.
	   *
	   * @throws IllegalArgumentException If there isn't enough information
	   *                                  to compare them.
	   */
	  public int compare(Object o1, Object o2)
	    {
	      DicomImage d1 = (DicomImage)o1;
	      DicomImage d2 = (DicomImage)o2;

	      // Determine if the Image Position coordinates exists
	      if ( d1.getX() != null && d1.getY() != null &&
		   d1.getZ() != null && d2.getX() != null && 
		   d2.getY() != null && d2.getZ() != null )
		{
		  double x1 = d1.getX().doubleValue();
		  double y1 = d1.getY().doubleValue();
		  double z1 = d1.getZ().doubleValue();
		  double x2 = d2.getX().doubleValue();
		  double y2 = d2.getY().doubleValue();
		  double z2 = d2.getZ().doubleValue();

		  // *****
		  //
		  // DICOM STANDARDS:
		  //
		  // The x-y-z coordinate system on a patient is defined so
		  // the x axis points to the patient's left, the y axis points
		  // to the patient's posterior, and the z axis points to the 
		  // patient's head:
		  //
		  //    (*+*)     ^ Z
		  //     _|_      |
		  //      |       | Y
		  //     / \      x -- > X
		  //
		  //
		  // TALAIRACH COORDINATES:
		  //
		  // The x-y-z coordinate system on a patient is defined so
		  // that the positive x axis points from the patient's left
		  // to right, the positive y axis points from posterior to
		  // anterior, and the positive z axis points from inferior
		  // to superior:
		  //
		  //    (*+*)            ^ Z
		  //     _|_             |
		  //      |            Y |
		  //     / \      X < -- 0
		  // 
		  //
		  // *****

		  double u1 = 0;
		  double u2 = 0;

		  // Sort in opposite x directions
		  if ( _sortingCoordinate.equals(Talairach.POSITIVE_X) ) {
		    u1 = x2;
		    u2 = x1;
		  }
		  if ( _sortingCoordinate.equals(Talairach.NEGATIVE_X) ) {
		    u1 = x1;
		    u2 = x2;
		  }

		  // Sort in opposite y directions
		  if ( _sortingCoordinate.equals(Talairach.POSITIVE_Y) ) {
		    u1 = y2;
		    u2 = y1;
		  }
		  if ( _sortingCoordinate.equals(Talairach.NEGATIVE_Y) ) {
		    u1 = y1;
		    u2 = y2;
		  }

		  // Sort in same z directions
		  if ( _sortingCoordinate.equals(Talairach.POSITIVE_Z) ) {
		    u1 = z1;
		    u2 = z2;
		  }
		  if ( _sortingCoordinate.equals(Talairach.NEGATIVE_Z) ) {
		    u1 = z2;
		    u2 = z1;
		  }

		  // Order along increasing coordinates
		  if (u1 < u2) { return -1; }
		  if (u1 == u2) { return 0; }
		  return 1;
		}

	      // Can't make the comparison
	      throw new IllegalArgumentException("Not enough information to " +
						 "sort DICOM images.");
	    }

	  /**
	   * Indicates whether some object is equal to this Comparator.
	   *
	   * @param obj The reference object with which to compare.
	   *
	   * @return True only if the specified object is also a comparator
	   *         and it imposes the same ordering as this comparator.
	   */
	  public boolean equals(Object obj)
	    {
	      return super.equals(obj);
	    }
	};
    }

  /** Class used to hold a Buffered Image with its metadata. */
  private class DicomImage
  {
    /** X coordinate of the first image pixel. */
    private Double _x;

    /** Y coordinate of the first image pixel. */
    private Double _y;

    /** Z coordinate of the first image pixel. */
    private Double _z;

    /** Buffered Image. */
    private BufferedImage _image;

    /**
     * Constructs a DicomImage.
     *
     * @param x X coordinate of the first image pixel.
     * @param y Y coordinate of the first image pixel.
     * @param z Z coordinate of the first image pixel.
     * @param image Buffered Image.
     */
    public DicomImage(Double x, Double y, Double z, BufferedImage image)
      {
	_x = x;
	_y = y;
	_z = z;
	_image = image;
      }

    /**
     * Determines if valid X,Y,Z coordinates are available.
     *
     * @return True if valid X,Y,Z coordinates are available; false otherwise.
     */
    public boolean hasValidCoordinates()
      {
	return (_x != null && _y != null && _z != null);
      }

    /**
     * Gets the X coordinate.
     *
     * @return X coordinate of the first image pixel.
     */
    public Double getX()
      {
	return _x;
      }

    /**
     * Gets the Y coordinate.
     *
     * @return Y coordinate of the first image pixel.
     */
    public Double getY()
      {
	return _y;
      }

    /**
     * Gets the Z coordinate.
     *
     * @return Z coordinate of the first image pixel.
     */
    public Double getZ()
      {
	return _z;
      }

    /**
     * Gets the Buffered Image.
     *
     * @return Buffered Image.
     */
    public BufferedImage getBufferedImage()
      {
	return _image;
      }
  }
}
