/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.talairach;

/**
 * Properties of Talairach space.
 *
 * @version 1 May 2003
 */
public class Talairach
{
  /** Positive X direction in Talairach space. */
  public static String POSITIVE_X = "+x";

  /** Negative X direction in Talairach space. */
  public static String NEGATIVE_X = "-x";

  /** Positive Y direction in Talairach space. */
  public static String POSITIVE_Y = "+y";

  /** Negative Y direction in Talairach space. */
  public static String NEGATIVE_Y = "-y";

  /** Positive Z direction in Talairach space. */
  public static String POSITIVE_Z = "+z";

  /** Negative Z direction in Talairach space. */
  public static String NEGATIVE_Z = "-z";

  /** Constructs a Talairach. */
  private Talairach()
    {
    }

  /**
   * Determines if the Talairach coordinate is valid.
   *
   * @param coordinate Talairach coordinate to check for validity.
   *
   * @return True if the Talairach coordinate is valid; false otherwise.
   */
  public static boolean isValid(String coordinate)
    {
      // Check for a null value
      if (coordinate == null) { return false; }

      // Determine if it equals a valid Talairach coordinate
      if ( !coordinate.equals(Talairach.POSITIVE_X) &&
	   !coordinate.equals(Talairach.NEGATIVE_X) &&
	   !coordinate.equals(Talairach.POSITIVE_Y) &&
	   !coordinate.equals(Talairach.NEGATIVE_Y) &&
	   !coordinate.equals(Talairach.POSITIVE_Z) &&
	   !coordinate.equals(Talairach.NEGATIVE_Z) )
	{
	  return false;
	}

      // Is valid
      return true;
    }

  /**
   * Determines if the Talairach coordinate has a positive sign.
   *
   * @param coordinate Talairach coordinate to determine the sign of.
   *
   * @return True if the Talairach coordinate is positive; false otherwise.
   */
  public static boolean isPositive(String coordinate)
    {
      // Look at the first character
      return coordinate.startsWith("+");
    }

  /**
   * Gets the name of the Talairach coordinate variable.
   *
   * @param coordinate Talairach coordinate to get the variable name of.
   *
   * @return Talairach coordinate variable name.
   */
  public static String getVariable(String coordinate)
    {
      // Return the second character
      return coordinate.substring(1, 2);
    }

  /**
   * Reverses the sign of the Talairach coordinate.
   *
   * @param coordinate Talairach coordinate to reverse the sign of.
   *
   * @return Talairach coordinate with its sign reversed.
   */
  public static String reverseSign(String coordinate)
    {
      if ( isPositive(coordinate) ) {
	return coordinate.replace('+', '-');
      }
      else {
	return coordinate.replace('-', '+');
      }
    }
}
