/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that transforms a point in Talairach space.
 *
 * @version 12 December 2003
 */
public class TalairachPointTransformer extends Processor
{
  /** Allowed Talairach space directions. */
  private static String DIRS = Talairach.POSITIVE_X + ", " + 
                               Talairach.NEGATIVE_X + ", " +
                               Talairach.POSITIVE_Y + ", " +
                               Talairach.NEGATIVE_Y + ", " +
                               Talairach.POSITIVE_Z + ", " +
                               Talairach.NEGATIVE_Z;

  /** Description of this Processor. */
  private static String DESC = "Transforms a point in Talairach space.";

  /** Node name for the Talairach coordinate for the image rows. */
  private static String ROW_NAME = "Row Talairach Coordinate";

  /** Node description for the row Talairach coordinate. */
  private static String ROW_DESC = DIRS;

  /** Node name for the Talairach coordinate for the image columns. */
  private static String COL_NAME = "Column Talairach Coordinate";

  /** Node description for the column Talairach coordinate. */
  private static String COL_DESC = DIRS;

  /** Node name for the Talairach coordinate for the image stack. */
  private static String STACK_NAME = "Stack Talairach Coordinate";

  /** Node description for the stack Talairach coordinate. */
  private static String STACK_DESC = DIRS;

  /** Input Node name for the row size. */
  private static String ROW_SIZE_NAME = "Row Size";

  /** Input Node description for the row size. */
  private static String ROW_SIZE_DESC =
  "Size of the images along the row direction.";

  /** Input Node name for the column size. */
  private static String COL_SIZE_NAME = "Column Size";

  /** Input Node description for the column size. */
  private static String COL_SIZE_DESC =
  "Size of the images along the column direction.";

  /** Input Node name for the stack size. */
  private static String STK_SIZE_NAME = "Stack Size";

  /** Input Node description for the stack size. */
  private static String STK_SIZE_DESC =
  "Size of the images along the stack direction.";

  /** Input Node name for the X coordinate of the point to transform. */
  private static String X_NAME = "Row Coordinate";

  /** Input Node description for the X coordinate of the point to transform. */
  private static String X_DESC = "Row coordinate of the point to transform.";

  /** Input Node name for the Y coordinate of the point to transform. */
  private static String Y_NAME = "Column Coordinate";

  /** Input Node description for the Y coordinate of the point to transform. */
  private static String Y_DESC =
  "Column coordinate of the point to transform.";

  /** Input Node name for the Z coordinate of the point to transform. */
  private static String Z_NAME = "Stack Coordinate";

  /** Input Node description for the Z coordinate of the point to transform. */
  private static String Z_DESC = "Stack coordinate of the point to transform.";

  /** Output Node name for the X coordinate of the transformed point. */
  private static String NEW_X_NAME = "New Row Coordinate";

  /** Output Node description for the X coordinate of the transformed point. */
  private static String NEW_X_DESC =
  "New row coordinate of the transformed point.";

  /** Output Node name for the Y coordinate of the transformed point. */
  private static String NEW_Y_NAME = "New Column Coordinate";

  /** Output Node description for the Y coordinate of the transformed point. */
  private static String NEW_Y_DESC =
  "New column coordinate of the transformed point.";

  /** Output Node name for the Z coordinate of the transformed point. */
  private static String NEW_Z_NAME = "New Stack Coordinate";

  /** Output Node description for the Z coordinate of the transformed point. */
  private static String NEW_Z_DESC =
  "New stack coordinate of the transformed point.";

  /**
   * Constructs a Talairach Point Transformer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachPointTransformer(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the current Talairach Coordinates
      Class type = String.class;
      InputNode inputNode = new InputNode(ROW_NAME, ROW_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(COL_NAME, COL_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(STACK_NAME, STACK_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the new Talairach Coordinates
      type = String.class;
      inputNode = new InputNode("New "+ROW_NAME, ROW_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode("New "+COL_NAME, COL_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode("New "+STACK_NAME, STACK_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the image sizes
      type = Double.class;
      inputNode = new InputNode(ROW_SIZE_NAME, ROW_SIZE_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_SIZE_NAME, COL_SIZE_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STK_SIZE_NAME, STK_SIZE_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the coordinates of the point to transform
      type = Double.class;
      inputNode = new InputNode(X_NAME, X_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(Y_NAME, Y_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(Z_NAME, Z_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Nodes for the transformed point coordinates
      type = Double.class;
      OutputNode node = new OutputNode(NEW_X_NAME, NEW_X_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(NEW_Y_NAME, NEW_Y_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(NEW_Z_NAME, NEW_Z_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Talairach Point Transformer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachPointTransformer(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input Node values for the current Talairach Coordinates
	String rowC = (String)_getInputValue(ROW_NAME, null);
	String colC = (String)_getInputValue(COL_NAME, null);
	String stkC = (String)_getInputValue(STACK_NAME, null);

	// Get the Input Node values for the new Talairach Coordinates
	String newRowC = (String)_getInputValue("New "+ROW_NAME, null);
	String newColC = (String)_getInputValue("New "+COL_NAME, null);
	String newStkC = (String)_getInputValue("New "+STACK_NAME, null);

	// Get the Input Node values for the image sizes
	Double rowSize = (Double)_getInputValue(ROW_SIZE_NAME, new Double(0));
	Double colSize = (Double)_getInputValue(COL_SIZE_NAME, new Double(0));
	Double stkSize = (Double)_getInputValue(STK_SIZE_NAME, new Double(0));

	// Get the Input Node values for the coordinates of the point
	Double xCoord = (Double)_getInputValue(X_NAME, null);
	Double yCoord = (Double)_getInputValue(Y_NAME, null);
	Double zCoord = (Double)_getInputValue(Z_NAME, null);

	// Not enough information
	if ( !Talairach.isValid(rowC) || !Talairach.isValid(colC) ||
	     !Talairach.isValid(stkC) || !Talairach.isValid(newRowC) ||
	     !Talairach.isValid(newColC) || !Talairach.isValid(newStkC) ||
	     xCoord == null || yCoord == null || zCoord == null )
	  {
	    return;
	  }

	// Set the initial values
	double rSize = rowSize.doubleValue();
	double cSize = colSize.doubleValue();
	double sSize = stkSize.doubleValue();
	double rVal = xCoord.doubleValue();
	double cVal = yCoord.doubleValue();
	double sVal = zCoord.doubleValue();

	// Translate the point coordinates
	double newRVal = _getTranslatedValue(newRowC, rowC, rSize, rVal, colC,
					     cSize, cVal, stkC, sSize, sVal);
	double newCVal = _getTranslatedValue(newColC, rowC, rSize, rVal, colC,
					     cSize, cVal, stkC, sSize, sVal);
	double newSVal = _getTranslatedValue(newStkC, rowC, rSize, rVal, colC,
					     cSize, cVal, stkC, sSize, sVal);

	// Set the output values
	_setOutputValue(NEW_X_NAME, new Double(newRVal));
	_setOutputValue(NEW_Y_NAME, new Double(newCVal));
	_setOutputValue(NEW_Z_NAME, new Double(newSVal));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the translated value of the specified coordinate.
   *
   * @param newCoord Name of the coordinate to find the translated value of.
   * @param oldRowCoord Name of the current row coordinate.
   * @param oldRowSize Size of the images along the current row coordinate.
   * @param oldRowValue Value of the current row coordinate.
   * @param oldColCoord Name of the current column coordinate.
   * @param oldColSize Size of the images along the current column coordinate.
   * @param oldColValue Value of the current column coordinate.
   * @param oldStkCoord Name of the current stack coordinate.
   * @param oldStkSize Size of the images along the current stack coordinate.
   * @param oldStkValue Value of the current stack coordinate.
   *
   * @return Translated value of the specified coordinate.
   */
  private double _getTranslatedValue(String newCoord, String oldRowCoord,
				     double oldRowSize, double oldRowValue,
				     String oldColCoord, double oldColSize,
				     double oldColValue, String oldStkCoord,
				     double oldStkSize, double oldStkValue)
    {
      String newCoordVar = Talairach.getVariable(newCoord);

      // Match the new and old coordinate variables
      String coord = oldRowCoord;
      double size = oldRowSize;
      double value = oldRowValue;

      if ( newCoordVar.equals( Talairach.getVariable(oldColCoord) ) ) {
	coord = oldColCoord;
	size = oldColSize;
	value = oldColValue;
      }

      if ( newCoordVar.equals( Talairach.getVariable(oldStkCoord) ) ) {
	coord = oldStkCoord;
	size = oldStkSize;
	value = oldStkValue;
      }

      // Subtract from the size if the coordinates are of opposite signs
      if ( !newCoord.equals(coord) ) { return size - value; }
      return value;
    }
}
