/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector; 

/**
 * Processor that performs Number binary operations.
 *
 * @author Keith R. Ouellette
 * 
 * @version 24 June 2003
 */
public class ArithmeticOperator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs arithmetic binary operations";

  /** Base Input Node name for the ith Number. */
  private static String LEFT_OPERAND_NAME = "Left Operand"; 
  
  /** Base Input Node name for the ith Number. */
  private static String RIGHT_OPERAND_NAME = "Right Operand";

  /** Input Node description for the ith Number. */
  private static String NUMBER_DESC = "Number argument.";

  /** Output Node name for the resulting Number. */
  private static String COMBINED_NAME = "Resulting Number.";

  /** Output Node description for the resulting Number. */
  private static String COMBINED_DESC = "The result of the binary operation.";

  /** Property for the operation. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "Binary operation to execute.";

  /** Supported Number operation. */
  private static String ADD = "+";
  
  /** Supported Number operation. */
  private static String SUBTRACT = "-";
  
  /** Supported Number operation. */
  private static String MULT = "*";
  
  /** Supported Number operation. */
  private static String DIVIDE = "/";
  
  /**
   * Constructs a ArithmeticOperator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArithmeticOperator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a ArithmeticOperator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArithmeticOperator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create left operand InputNode
      name = LEFT_OPERAND_NAME;
      InputNode inputNode = new InputNode(name, NUMBER_DESC, Number.class);
      _addInputNode(inputNode);
      
      // Create right operand InputNode
      name = RIGHT_OPERAND_NAME;
      inputNode = new InputNode(name, NUMBER_DESC, Number.class);
      _addInputNode(inputNode);

      // Create the Output Node for the result of the operation
      OutputNode outputNode = new OutputNode(COMBINED_NAME, COMBINED_DESC,
					     Number.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the left operand
	Number numLeft 
	           = (Number)_getInputValue(LEFT_OPERAND_NAME,new Long(0));
	// Get the right operand
	Number numRight 
	           = (Number)_getInputValue(RIGHT_OPERAND_NAME, new Long(1)); 
	
	Number answer = null;

	// Get the Number operation
	String op = (String) _getPropertyValue(OP_PROP_NAME);

	if(op.equals(ADD)) {	  
	  answer = _add(numLeft, numRight);
	}
	else if (op.equals(SUBTRACT)) {
	  answer = _subtract(numLeft, numRight);
	}
	else if (op.equals(MULT)) {
	 answer = _multiply(numLeft, numRight);
	}
	else if (op.equals(DIVIDE)) {
	 answer = _divide(numLeft, numRight);
	}
		
	// Set the reulting value
	_setOutputValue(COMBINED_NAME, answer);
      }
      catch (ProcessingException e) {
	// Set the resulting value to null
	_setOutputValue(COMBINED_NAME, null);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /** 
   * Returns either a Long or Double with the same value as the given Number.
   * 
   * @return Long if n is either a Byte, Short, Integer, or Long.  
   * Double otherwise 
   *
   * @throws NullPointerException If n is null.
   */
  private Number _normalizeNumber(Number n)
    {
      if(n == null) {
	String msg = "Cannot have a null n";
	throw new NullPointerException(msg);
      }
      
      if(n instanceof Byte || 
	 n instanceof Short ||
	 n instanceof Integer ||	   
	 n instanceof Long) {
	return new Long(n.longValue());
      }
  
      return new Double(n.doubleValue());    
    }

  /** 
   * Returns the sum of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _add(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
      
      if(numLeft instanceof Long && numRight instanceof Long) {	
	long sum = numLeft.longValue() + numRight.longValue();
	return new Long(sum);
      }
      else {
	double sum = numLeft.doubleValue() + numRight.doubleValue();
	return new Double(sum);
      }   
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _subtract(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      if(numRight instanceof Long) {	
	Number negNumRight = new Long( - numRight.longValue()); 
	return _add(numLeft, negNumRight);
      }
      else {
	Number negNumRight = new Double( - numRight.doubleValue()); 
	return _add(numLeft, negNumRight);
      }   
    }

  /** 
   * Returns the difference of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _multiply(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      if(numLeft instanceof Long && numRight instanceof Long) {	
	long product = numLeft.longValue() * numRight.longValue();
	return new Long(product);
      }
      else {
	double product = numLeft.doubleValue() * numRight.doubleValue();
	return new Double(product);
      } 
    }

  /** 
   * Returns the quotient of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _divide(Number numLeft, Number numRight) 
    throws ProcessingException
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
	
      double right = numRight.doubleValue();
	
      // if(right == 0.0) {
      //  String msg = "Cannot divide by zero.";
      //   throw new ProcessingException(msg);
      // }
      
      Number invNumRight = new Double( 1.0 / right); 
      return _multiply(numLeft, invNumRight);   
    } 

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = ADD;

      // Create the Vector of supported binary operations.
      Vector ops = new Vector();
      ops.add(ADD);
      ops.add(SUBTRACT);
      ops.add(MULT);
      ops.add(DIVIDE);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable binary operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
