/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;
import java.util.Vector;

/**
 * Processor that adds elements to the end of an array.
 *
 * @author Scott C. Neu 
 *
 * @version 3 June 2003
 */
public class ArrayMaker extends Processor implements Stateful
{
  /** Description of this Processor. */
  private static String DESC = "Adds elements to the end of an array.";

  /** Input Node Name for the array element. */
  private static String ELEMENT_NAME = "Array Element";

  /** Input Node Description for the array element. */
  private static String ELEMENT_DESC =
  "Element that added to the end of the array.";

  /** Output Node Name for the array. */
  private static String ARRAY_NAME = "Array";

  /** Output Node Description for the array. */
  private static String ARRAY_DESC = "Array of elements.";

  /** Property for the array type. */
  private static String PROP_NAME = "type";

  /** Property description for the array type. */
  private static String PROP_DESC = "Type of array to make.";

  /** Elements in the array. */
  private Vector _elements;

  /**
   * Constructs an Array Maker with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArrayMaker(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an Array Maker using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArrayMaker(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the array type
      Class arrayType = (Class)_getPropertyValue(PROP_NAME);

      // Create the Input Node
      InputNode node = new InputNode(ELEMENT_NAME, ELEMENT_DESC, arrayType);
      _addInputNode(node);

      // Create the Output Node
      Class type = Array.newInstance(arrayType, 0).getClass();
      OutputNode outputNode = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addOutputNode(outputNode);

      // Initialize the state
      reinitializeState();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the element
	Object element = _getInputValue(ELEMENT_NAME, null);
	_elements.add(element);

	// Create a new array
	Class arrayType = (Class)_getPropertyValue(PROP_NAME);
	Object array = Array.newInstance(arrayType, _elements.size());

	// Add the elements to the new array
	for (int i = 0; i < _elements.size(); i++) {
	  Array.set(array, i, _elements.elementAt(i));
	}

	// Set the array
	_setOutputValue(ARRAY_NAME, array);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Reinitializes the state.  This is called by a Loop Module before a loop is
   * started.
   */
  public void reinitializeState()
    {
      _elements = new Vector();
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default array type
      String name = PROP_NAME;
      String desc = PROP_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
