/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that checks if a given Class represents an interface, array, 
 * or primitive.
 * 
 * @author Keith R. Ouellette
 * 
 * @version 3 June 2003
 */
public class ClassDeterminer extends Processor
{
  /** Description of this Processor. */
  private static String DESC 
                = "Checks if a given class represents an interface, array, " +
                    "or a primitive.";
 
  /** Input Node name for the Object to test. */
  private static String CLASS_NAME = "class";

  /** Input Node description for the object argument. */
  private static String CLASS_DESC = "Class argument";

  /** Output Node name for the Boolean. */
  private static String BOOLEAN_NAME = "answer";

  /** Output Node description for the Boolean. */
  private static String BOOLEAN_DESC = "Does class represent ";

  /** Property for the number of Numbers. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the number of Number. */
  private static String OP_PROP_DESC = "Operation to execute.";

  /** Supported Class operation. */
  private static String IS_ARRAY = "is array?"; 

  /** Supported Class operation. */
  private static String IS_INTERFACE = "is interface?";
  
  /** Supported Class operation. */
  private static String IS_PRIMITIVE = "is primitive?";
  
  /**
   * Constructs a ClassDeterminer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ClassDeterminer (String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an ClassDeterminer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ClassDeterminer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);
       	
      // Create the InputNode for the class argument
      InputNode inputNode = new InputNode(CLASS_NAME, CLASS_DESC,Class.class);
      _addInputNode(inputNode);

      // Get the Class operation
      String op = (String) _getPropertyValue(OP_PROP_NAME);
       
      String type = "";

      if(op.equals(IS_ARRAY)) {	  
        type = "an array?";
      }
      else if(op.equals(IS_INTERFACE)){
	type = "an interface?";	
      }
      else if(op.equals(IS_PRIMITIVE)){
	type = "a primtitive?";
      }
      String d = BOOLEAN_DESC + type;

      // Create the Output Node for the Boolean return value
      OutputNode outputNode = new OutputNode(BOOLEAN_NAME, d, Boolean.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	Boolean bool = null;

	// Get the Class to test.
	Class c = (Class) _getInputValue(CLASS_NAME, null);
	
	// Get the Class operation
	String op = (String) _getPropertyValue(OP_PROP_NAME);

	if(op.equals(IS_ARRAY)) {	  
	   bool = new Boolean(c.isArray());
	}
	else if(op.equals(IS_INTERFACE)){
	  bool = new Boolean(c.isInterface());
	}
	else if(op.equals(IS_PRIMITIVE)){
	  bool = new Boolean(c.isPrimitive());
	}
	// Set the combined String
	_setOutputValue(BOOLEAN_NAME, bool);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = IS_ARRAY;

      // Create the Vector of supported binary operations.
      Vector ops = new Vector();
      ops.add(IS_ARRAY);
      ops.add(IS_INTERFACE);
      ops.add(IS_PRIMITIVE);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable Class operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
