/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

/**
 * Processor that operates search methods on a Collection.
 *
 * @author Keith R. Ouellette 
 *
 * @version 5 June 2003
 */
public class CollectionsSearcher extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Operates search methods on a Collection.";

  /** Input Node name of the Collection to search. */
  private static String COLLECTION_NAME = "collection";

  /** Input Node description of the Collection to search. */
  private static String COLLECTION_DESC = "collection to search.";

  /** Input Node name of the method argument. */
  private static String KEY_NAME = "key";

  /** Input Node description of the method argument. */
  private static String KEY_DESC = "method argument.";
  
  /** Output Node name of the Object returned. */
  private static String OBJECT_NAME = "returned value";
 
  /** Property for the operation. */
  private static String OP_PROP_NAME = "search operation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "search operation to execute.";
  
  /** Collection operation. */
  private static String MAX = "max";
  
  /** Collection operation. */
  private static String MIN = "min"; 

  /** Collection operation. */
  private static String BINARY_SEARCH = "binary search"; 

  /**
   * Constructs a CollectionsSearcher with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public CollectionsSearcher(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a CollectionsSearcher using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public CollectionsSearcher(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the String operation 
      String op = (String)_getPropertyValue(OP_PROP_NAME);
     
      Class type = Collection.class;
      InputNode inputNode 
	   = new InputNode(COLLECTION_NAME, COLLECTION_DESC, type);
      
      if(op.equals(MAX)) {
	// Create the "Collection to search" Input Node
	_addInputNode(inputNode); 

	// Create the Output Node
	type = Object.class;
	OutputNode outputNode = new OutputNode(OBJECT_NAME, 
					 "Max value in the collection", type);
	_addOutputNode(outputNode);
      }
      else if(op.equals(MIN)) {
	// Create the "Collection to search" Input Node
	_addInputNode(inputNode);

	// Create the Output Node
	type = Object.class;
	OutputNode outputNode = new OutputNode(OBJECT_NAME, 
					 "Min value in the collection", type);
	_addOutputNode(outputNode);
      }
      else if(op.equals(BINARY_SEARCH)) {
	// Create the "List to search" Input Node
	type = List.class;
	inputNode = new InputNode(COLLECTION_NAME, COLLECTION_DESC, type);
	_addInputNode(inputNode);
	
	// Create the "search key" Input Node
	type = Object.class;
	inputNode = new InputNode(KEY_NAME, KEY_DESC, type);
	_addInputNode(inputNode); 

	// Create the Output Node
	type = Integer.class;
	OutputNode outputNode = new OutputNode(OBJECT_NAME, 
				 "Index of the key in the List", type);
	_addOutputNode(outputNode);	
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the Collection to search
	Object value = _getInputValue(COLLECTION_NAME, null);

	// Get the String operation 
	String op = (String)_getPropertyValue(OP_PROP_NAME);
       	
	if(op.equals(MAX)) {	
	  Collection c = (Collection) value;
	  Object max = Collections.max(c);
	  _setOutputValue(OBJECT_NAME, max);
	}
	else if(op.equals(MIN)) {	
	  Collection c = (Collection) value;
	  Object min = Collections.min(c);
	  _setOutputValue(OBJECT_NAME, min);
	}
	else if(op.equals(BINARY_SEARCH)) {	
	  List l = (List) value; 
	  Object key = _getInputValue(KEY_NAME, null);
	  int index = Collections.binarySearch(l, key);
	  _setOutputValue(OBJECT_NAME, new Integer(index));
	}		
	_setOutputValue(OBJECT_NAME, null);	
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = BINARY_SEARCH;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(MAX);
      ops.add(MIN);
      ops.add(BINARY_SEARCH);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable collection search operations " + 
	                  "for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
