/*
COPYRIGHT NOTICE
Copyright (c) 2002  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.util.Vector;

/**
 * Processor that returns a default Object if no valid Object is passed to it.
 *
 * @author Keith R. Ouellette
 * 
 * @version 5 June 2003
 */
public class Defaulter extends Processor
{ 
  /** Description of this Processor. */
  private static String DESC = "Returns a default if there's no valid input.";
  
  /** Input Node Name for the Object to check. */
  private static String IN_NAME = "Object";

  /** Input Node Description for the Object to check. */
  private static String IN_DESC = "Object to check.";

  /** Property for the default value. */
  private static String DEF_VALUE_NAME = "value";

  /** Property description for the default value. */
  private static String DEF_VALUE_DESC = "Default value.";
  
  /** Output Node Name for the returned Object. */
  private static String OUT_NAME = "Validated Object";

  /** Output Node Description for the returned Object. */
  private static String OUT_DESC = "Original Object or the default Object.";
 
  /** Property for the null value type. */
  private static String PROP_TYPE_NAME = "default type";

  /** Property description for the null value type. */
  private static String PROP_TYPE_DESC = "Type of default value."; 

  /**
   * Constructs a Defaulter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Defaulter(String name)
    {
      this(name, _getDefaultProperties());
    }
 
  /**
   * Constructs a Defaulter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Defaulter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      try {
	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);
	
	// Get the default value type and description
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
	
	// Support 1-D array
	String className = type.getName();
	
	if ( type.isArray() ) {
	  // Get the Class of the array
	  type = Class.forName( className.substring(2, name.length()) );
	  // Get a 1-D array of the Class
	  type = Array.newInstance(type, 0).getClass();
	}
	// Create left operand InputNode
	InputNode inputNode = new InputNode(IN_NAME, IN_DESC, type);
	_addInputNode(inputNode);
	
	// Create left operand InputNode
	inputNode = new InputNode(DEF_VALUE_NAME,DEF_VALUE_DESC,type);
	_addInputNode(inputNode);
	
	// Create the Output Node for the result of the operation
	OutputNode outputNode = new OutputNode(OUT_NAME, OUT_DESC, type);
	_addOutputNode(outputNode); 
      }
      catch (Exception e) {
	String msg = "Unable to create \"" + this + "\".";
	throw new IllegalArgumentException(msg);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default value type
      String name = PROP_TYPE_NAME;
      String desc = PROP_TYPE_DESC;
      Object value = String.class;
      properties[0] = new ProcessorProperty(name, desc, value);

      return properties;
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the default value
	Object defaultValue = _getInputValue(DEF_VALUE_NAME, null);
	// Get the input Object
	Object value = _getInputValue(IN_NAME, defaultValue);
	// Get the default value type
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
	
	if ( value != null && !type.isInstance(value) ) {
	  // Assign the default value
	  value = defaultValue;
	}
	// Set the returned Object
	_setOutputValue(OUT_NAME, value);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
