/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that performs the Java instanceof Operation.
 *
 * @author Keith R. Ouellette
 * 
 * @version 3 June 2003
 */
public class InstanceofOperator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs the Java instanceof Operation.";
 
  /** Input Node name for the Object to test. */
  private static String OBJECT_NAME = "object";

  /** Input Node description for the object argument. */
  private static String OBJECT_DESC = "Object argument";

  /** Input Node name for the Object to test. */
  private static String CLASS_NAME = "class";

  /** Input Node description for the object argument. */
  private static String CLASS_DESC = "Class argument";

  /** Output Node name for the Boolean. */
  private static String BOOLEAN_NAME = "instanceof?";

  /** Output Node description for the Boolean. */
  private static String BOOLEAN_DESC = "Is object instanceof class?";

  /**
   * Constructs a InstanceofOperator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public InstanceofOperator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a InstanceofOperator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public InstanceofOperator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node for the object argument
      InputNode inputNode 
	     = new InputNode(OBJECT_NAME, OBJECT_DESC, Object.class);
      _addInputNode(inputNode);
       
      // Create the InputNode for the class argument
      inputNode = new InputNode(CLASS_NAME, CLASS_DESC, Class.class);
      _addInputNode(inputNode);

      // Create the Output Node for the Boolean return value
      OutputNode outputNode = new OutputNode(BOOLEAN_NAME, 
					     BOOLEAN_DESC,
					     Boolean.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the Object to test.
	Object o = _getInputValue(OBJECT_NAME, null);

	// Get the Class to test.
	Class c = (Class) _getInputValue(CLASS_NAME, null);
	
	boolean b = c.isInstance(o);

	// Set the combined String
	_setOutputValue(BOOLEAN_NAME, new Boolean(b));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[0];
      return properties;
    }
}
