/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that performs Byte, Short, Integer, or Long binary operations.
 *
 * @author Keith R. Ouellette
 * 
 * @version 3 June 2003
 */
public class ModAndBitwiseOperator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs binary operations on Bytes," 
                                + " Shorts, Integers, or Longs";

  /** Base Input Node name for left operand. */
  private static String LEFT_OPERAND_NAME = "Left Operand"; 
  
  /** Base Input Node name for the right opeand. */
  private static String RIGHT_OPERAND_NAME = "Right Operand";

  /** Input Node description for the ith operand. */
  private static String NUMBER_DESC = "Number argument.";

  /** Output Node name for the resulting Number. */
  private static String COMBINED_NAME = "Resulting Number.";

  /** Output Node description for the resulting Number. */
  private static String COMBINED_DESC = "The result of the binary operation.";

  /** Property for the number of Numbers. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the number of Number. */
  private static String OP_PROP_DESC = "Binary operation to execute.";

  /** Property for the number of Numbers. */
  private static String LEFT_OP_PROP_NAME = "left operand type";

  /** Property description for the number of Number. */
  private static String LEFT_OP_PROP_DESC = "Class type of the left operand.";
  
  /** Property for the number of Numbers. */
  private static String RIGHT_OP_PROP_NAME = "right operand type";

  /** Property description for the number of Number. */
  private static String RIGHT_OP_PROP_DESC ="Class type of the right operand.";
  
  /** Supported Number operation. */
  private static String MOD = "%";

  /** Supported Number operation. */
  private static String BITWISE_AND = "&";

  /** Supported Number operation. */
  private static String BITWISE_OR = "|";
  
  /** Supported Number operation. */
  private static String BITWISE_XOR = "^";

  /** Supported Number operation. */
  private static String BIT_SHIFT_LEFT = "<<";
  
  /** Supported Number operation */
  private static String BIT_SHIFT_RIGHT = ">>";

  /** 
   * Supported Long operation.
   * Shifts the bits in its left operand to the right by the number of
   * bits specified in its right operand- 0's are shifted in from the left.
   */
  private static String BIT_SHIFT_RIGHT_WITH_ZERO_EXTENSION = ">>>";
  
  /**
   * Constructs a ModAndBitwiseOperator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ModAndBitwiseOperator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a ModAndBitwiseOperator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ModAndBitwiseOperator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the left operand type
      Class leftType = (Class) _getPropertyValue(LEFT_OP_PROP_NAME);
      
      // Create left operand InputNode
      name = LEFT_OPERAND_NAME;
      InputNode inputNode = new InputNode(name, NUMBER_DESC, leftType);
      _addInputNode(inputNode);
      
      // Get the right operand type
      Class rightType = (Class) _getPropertyValue(RIGHT_OP_PROP_NAME);
      
      // Create right operand InputNode
      name = RIGHT_OPERAND_NAME;
      inputNode = new InputNode(name, NUMBER_DESC, rightType);

      // Get the Number operation
      String op = (String) _getPropertyValue(OP_PROP_NAME);

      // Add a positive range restriction on the right operand for bit shift 
      if (op.equals(BIT_SHIFT_LEFT) || 
	  op.equals(BIT_SHIFT_RIGHT) ||
	  op.equals(BIT_SHIFT_RIGHT_WITH_ZERO_EXTENSION)) {
       String n = "Positive Numbers";
       String d = "Only numbers greater than or equal to zero are allowed.";
       Restriction r = new RangeRestriction(n, d, 0, Long.MAX_VALUE);
       inputNode.setRestriction(r);
      }
      _addInputNode(inputNode);
 
      // Create the Output Node for the result of the operation
      OutputNode outputNode = new OutputNode(COMBINED_NAME, COMBINED_DESC,
					     Number.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the left operand
	Number numLeft 
	           = (Number)_getInputValue(LEFT_OPERAND_NAME,new Long(0));
	// Get the right operand
	Number numRight 
	           = (Number)_getInputValue(RIGHT_OPERAND_NAME, new Long(1)); 
	
	Number answer = null;

	// Get the Number operation
	String op = (String) _getPropertyValue(OP_PROP_NAME);

	if (op.equals(MOD)) {
	 answer = _mod(numLeft, numRight);
	}
	else if (op.equals(BITWISE_AND)) {
	 answer = _bitwiseAnd(numLeft, numRight);
	}
	else if (op.equals(BITWISE_OR)) {
	 answer = _bitwiseOr(numLeft, numRight);
	}
	else if (op.equals(BITWISE_XOR)) {
	 answer = _bitwiseXOr(numLeft, numRight);
	}
	else if (op.equals(BIT_SHIFT_LEFT)) {
	 answer = _bitShiftLeft(numLeft, numRight);
	}
	else if (op.equals(BIT_SHIFT_RIGHT)) {
	 answer = _bitShiftRight(numLeft, numRight);
	}
	else if (op.equals(BIT_SHIFT_RIGHT_WITH_ZERO_EXTENSION)) {
	  answer = _bitShiftRightWithZeroExtension(numLeft, numRight);
	}
		
	// Set the value of the operation
	_setOutputValue(COMBINED_NAME, answer);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /** 
   * Returns either a Long or Double with the same value as the given Number.
   * 
   * @return Long if n is either a Byte, Short, Integer, or Long.  
   * Double otherwise 
   *
   * @throws NullPointerException If n is null.
   */
  private Number _normalizeNumber(Number n)
    {
      if(n == null) {
	String msg = "Cannot have a null n";
	throw new NullPointerException(msg);
      }
      
      return new Long(n.longValue());       
    }

  /** 
   * Returns the value of numLeft mod numRight.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _mod(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);

      return new Long(numLeft.longValue() % numRight.longValue());   
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitwiseAnd(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      return new Long(numLeft.longValue() & numRight.longValue());
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitwiseOr(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      return new Long(numLeft.longValue() | numRight.longValue());
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitwiseXOr(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      return new Long(numLeft.longValue() ^ numRight.longValue());
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitShiftLeft(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
   
      return new Long(numLeft.longValue() << numRight.longValue());
    } 

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitShiftRight(Number numLeft, Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
    
      return new Long(numLeft.longValue() >> numRight.longValue());
    }

  /** 
   * Returns the product of the given numbers.
   *
   * @throws NullPointerException If numLeft or numRight is null.
   */
  private Number _bitShiftRightWithZeroExtension(Number numLeft, 
						 Number numRight)
    {
      numLeft = _normalizeNumber(numLeft);
      numRight = _normalizeNumber(numRight);
    
      return new Long(numLeft.longValue() >>> numRight.longValue());
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = MOD;

      // Create the Vector of supported binary operations.
      Vector ops = new Vector();
      ops.add(MOD);
      ops.add(BITWISE_AND);
      ops.add(BITWISE_OR);
      ops.add(BITWISE_XOR);
      ops.add(BIT_SHIFT_LEFT);
      ops.add(BIT_SHIFT_RIGHT);
      ops.add(BIT_SHIFT_RIGHT_WITH_ZERO_EXTENSION);
  
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable boolean operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
     
      // Create the Vector of supported operand types.
      Vector options = new Vector();
      options.add(Byte.class);
      options.add(Short.class);
      options.add(Integer.class);
      options.add(Long.class);
       
      // Create a Restriction to limit the allowed supported operand types
      REST_NAME = "Allowable operand types";
      REST_DESC = "Allowable input node types for this processor";
      restriction = new SetRestriction(REST_NAME, REST_DESC, options);
      
      // Create a ProcessorProperty for left operand type
      name = LEFT_OP_PROP_NAME;
      desc = LEFT_OP_PROP_DESC;
      value = Long.class;
      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      // Create a ProcessorProperty for right operand type
      name = RIGHT_OP_PROP_NAME;
      desc = RIGHT_OP_PROP_DESC;
      value = Long.class;
      properties[2] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
