/*
COPYRIGHT NOTICE
Copyright (c) 2002  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.exceptions.ProcessingException;

/**
 * Processor that returns the negation of a boolean.
 *
 * @author Scott C. Neu 
 *
 * @version 3 June 2003
 */
public class Not extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Negates a boolean value.";

  /** Input Node Name for the Boolean to negate. */
  private static String BOOL_NAME = "Boolean";

  /** Input Node Description for the Boolean to negate. */
  private static String BOOL_DESC = "Boolean to negate.";

  /** Output Node Name for the negated Boolean. */
  private static String NOT_NAME = "Negated Boolean";

  /** Output Node Description for the negated Boolean. */
  private static String NOT_DESC = "Boolean that is negated.";

  /**
   * Constructs a Not with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Not(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Not using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Not(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);
 
      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node for the Boolean to negate
      Class type = Boolean.class;
      InputNode inputNode = new InputNode(BOOL_NAME, BOOL_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the negated Boolean
      type = Boolean.class;
      OutputNode node = new OutputNode(NOT_NAME, NOT_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the input value
	Boolean bool = null;
	Object value = _getInputValue(BOOL_NAME, new Boolean(false) );
	if (value instanceof Boolean) { bool = (Boolean)value; }

	// Negate the Boolean if it is present
	Boolean notBool = null;
	if (bool != null) {
	  notBool = new Boolean( !bool.booleanValue() );
	}

	// Set the negated Boolean
	_setOutputValue(NOT_NAME, notBool);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[0];

      return properties;
    }
}
