/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * Processor that performs utility operations on a String.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class StringUtilities extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Operates on the given a String with " +
                               "the user specified utility methods.";

  /** Input Node name of the String to check. */
  private static String STRING_NAME = "string";

  /** Input Node description of the String to check. */
  private static String STRING_DESC = "String to operate on.";

  /** Input Node name of the String argument. */
  private static String ARG_NAME = "arg";

  /** Input Node description of the String argument. */
  private static String ARG_DESC = "String argument.";

  /** Input Node name of the Boolean argument. */
  private static String ARG2_NAME = "arg 2";
  
  /** Output Node name of the Integer returned. */
  private static String INT_NAME = "returned value";

  /** Output Node description of the Integer returned. */
  private static String INT_DESC = "Integer returned according to the method.";
 
  /** Output Node name of the Integer returned. */
  private static String TOKENS_NAME = "returned tokens";

  /** Output Node description of the Integer returned. */
  private static String TOKENS_DESC = "Tokens of the String returned.";
  
  /** Property for the operation. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "String operation to execute.";
  
  /** String operation. */
  private static String INDEX_OF = "index where expression first occurs";
  
  /** String operation. */
  private static String LENGTH = "string length";

  /** String operation. */
  private static String TOKENIZE = "tokenize"; 

  /**
   * Constructs a StringUtilities with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringUtilities(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a StringUtilities using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringUtilities(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the String operation 
      String op = (String)_getPropertyValue(OP_PROP_NAME);
      
      // Create the String argument Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);
      
      if(op.equals(LENGTH)) { 
	// Create the Output Node
	type = Integer.class;
	OutputNode outputNode = new OutputNode(INT_NAME, INT_DESC, type);
	_addOutputNode(outputNode);
      }
      if(op.equals(INDEX_OF)) {
	// Create the "SubString to check for" Input Node
	type = String.class;
	inputNode = new InputNode(ARG_NAME, ARG_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node
	type = Integer.class;
	OutputNode outputNode = new OutputNode(INT_NAME, INT_DESC, type);
	_addOutputNode(outputNode);
      }
      else if(op.equals(TOKENIZE)) {
	// Create the "Token delimiters" Input Node
	type = String.class;
	inputNode = new InputNode(ARG_NAME, "Token delimiters", type);
	_addInputNode(inputNode);

	// Create the "Return delimiters as tokens" Input Node
	type = Boolean.class;
	inputNode = new InputNode(ARG2_NAME, "Return delimiters as tokens?", 
				  type);
	_addInputNode(inputNode);

	// Create the Output Node
	type = String[].class;
	OutputNode outputNode = new OutputNode(TOKENS_NAME, TOKENS_DESC, type);
	_addOutputNode(outputNode);
      }     
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Create a reference to the return value
	Integer returnVal = null;

	// Get the String to check
	String value = (String)_getInputValue(STRING_NAME, "");

	// Get the String operation 
	String op = (String)_getPropertyValue(OP_PROP_NAME);
       	
	if(op.equals(INDEX_OF)) {	
	  // Get the String argument
	  String arg = (String)_getInputValue(ARG_NAME, "");
	  returnVal = new Integer(value.indexOf(arg));
	  _setOutputValue(INT_NAME, returnVal);
	}
	else if(op.equals(LENGTH)) {	
	  returnVal = new Integer(value.length());
	  _setOutputValue(INT_NAME, returnVal);
	}
	else if(op.equals(TOKENIZE)) {

	  Vector tokens = new Vector();
	  String delims = (String) _getInputValue(ARG_NAME, "\n");
	  Boolean returnDelims 
	    = (Boolean) _getInputValue(ARG2_NAME, new Boolean(false));
	  
	  StringTokenizer st = new StringTokenizer(value, delims,   
						 returnDelims.booleanValue());
	  while(st.hasMoreTokens()) {
	    tokens.add(st.nextToken());
	  }
	  String token[] = new String[tokens.size()];
	  for(int i = 0; i < token.length; i++) {
	    token[i] = (String) tokens.elementAt(i);
	  }
	  _setOutputValue(TOKENS_NAME, token);
	}		
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = INDEX_OF;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(INDEX_OF);
      ops.add(LENGTH);
      ops.add(TOKENIZE);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable String operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
