/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Vector;

/**
 * Stateful Processor that holds values inside a Loop Module.
 *
 * @author Scott C. Neu 
 *
 * @version 3 June 2003
 */
public class ValueHolder extends Processor implements Stateful
{
  /** Description of this Processor. */
  private static String DESC = "Holds values inside a loop.";

  /** Input Node Name for the value to store. */
  private static String VALUE_NAME = "Value";

  /** Input Node Description for the value to store. */
  private static String VALUE_DESC = "Value to store.";

  /** Input Node Name for the index of the value to store. */
  private static String INDEX_NAME = "Value Index";

  /** Input Node Description for the index of the value to store. */
  private static String INDEX_DESC = "Index of the value to store.";

  /** Base Output Node Name for the stored values. */
  private static String BASE_NAME = "Stored Value #";

  /** Base Output Node Description for the stored values. */
  private static String BASE_DESC = "Value stored at index ";

  /** Property for the stored value type. */
  private static String PROP_VALUE_NAME = "type";

  /** Property description for the stored value type. */
  private static String PROP_VALUE_DESC = "Type of value to store.";

  /** Property for the number of stored values. */
  private static String PROP_NUMBER_NAME = "number";

  /** Property description for the number of stored values. */
  private static String PROP_NUMBER_DESC = "Number of values to store.";

  /** Stored values. */
  private Vector _storedValues;

  /**
   * Constructs a Value Holder with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ValueHolder(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Value Holder using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ValueHolder(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the number and type of stored values
      Class valueType = (Class)_getPropertyValue(PROP_VALUE_NAME);
      int number = ((Integer)_getPropertyValue(PROP_NUMBER_NAME)).intValue();

      // Create the Input Node for the value to store
      InputNode inputNode = new InputNode(VALUE_NAME, VALUE_DESC, valueType);
      _addInputNode(inputNode);

      // Create the Input Node for the index of the value to store
      Class type = Integer.class;
      inputNode = new InputNode(INDEX_NAME, INDEX_DESC, type);
      _addInputNode(inputNode);

      // Create an Output Node for each stored value
      for (int i = 0; i < number; i++) {
	name = BASE_NAME + i;
	String desc = BASE_DESC + i;
	OutputNode outputNode = new OutputNode(name, desc, valueType);
	_addOutputNode(outputNode);
      }

      // Clear the stored values
      reinitializeState();
    }

  /**
   * Reinitializes the state.  This is called by a Loop Module before a loop is
   * started.
   */
  public void reinitializeState()
    {
      int number = ((Integer)_getPropertyValue(PROP_NUMBER_NAME)).intValue();
      _storedValues = new Vector(number);
      _storedValues.setSize(number);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the index of the value to store
	Integer indexValue = (Integer)_getInputValue(INDEX_NAME,
						     new Integer(0));
	int index = indexValue.intValue();
	
	// Store the value at the index
	Object value = _getInputValue(VALUE_NAME, null);
	_storedValues.setElementAt(value, index);

	// Set the Output Nodes
	for (int i = 0; i < _storedValues.size(); i++) {
	  _setOutputValue(BASE_NAME + i, _storedValues.elementAt(i));
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[2];

      // Default stored value type
      String name = PROP_VALUE_NAME;
      String desc = PROP_VALUE_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Number of stored values
      name = PROP_NUMBER_NAME;
      desc = PROP_NUMBER_DESC;
      value = new Integer(1);
      restriction = Utilities.getPositiveRestriction();
      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
