/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that replaces a String with another with respect to the user 
 * specified criteria.
 *
 * @author Bryan C. Vann 
 *
 * @version 30 June 2003
 */
public class FileNameReplacer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Replaces a Filename with another with "+
                               "respect to the user specified criteria.";

  /** Input Node name of the String to replace. */
  private static String FILENAME_NAME = "filename";

  /** Input Node description of the String to replace. */
  private static String FILENAME_DESC = "filename to operate on"; 

  /** Input Node name of the String argument. */
  private static String ARG1_NAME = "old extension";

  /** Input Node description of the String argument. */
  private static String ARG1_DESC = "File extension on input node.";

  /** Input Node name of the String argument. */
  private static String ARG2_NAME = "new extension";

  /** Input Node description of the String argument. */
  private static String ARG2_DESC = "File extension to replace old with.";
  
  /** Output Node name of the Replacement String. */
  private static String REP_STRING_NAME = "result";

  /** Output Node description of the Replacement String. */
  private static String REP_STRING_DESC = "string returned.";
  
  /** Property for matching case. */
  private static String MATCH_PROP_NAME = "match case";

  /** Property description for the operation. */
  private static String MATCH_PROP_DESC = "Specifies matching extension case.";

  /** Case Matching criteria. */
  private static String TRUE = "true";

  /** Case Matching criteria. */
  private static String FALSE = "false";

  /**
   * Constructs a FileNameReplacer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public FileNameReplacer(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a FileNameReplacer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public FileNameReplacer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the File Name Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(FILENAME_NAME, FILENAME_DESC, type);
      _addInputNode(inputNode);

      // Create the Old Extension Input Node
      type = String.class;
      inputNode = new InputNode(ARG1_NAME, ARG1_DESC, type);
      _addInputNode(inputNode);

      // Create the New Extension Input Node
      type = String.class;
      inputNode = new InputNode(ARG2_NAME, ARG2_DESC, type);
      _addInputNode(inputNode);
      
      // Create the Output Node
      type = String.class;
      OutputNode outputNode = new OutputNode(REP_STRING_NAME, 
					     REP_STRING_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the String operation 
	String matchString = (String)_getPropertyValue(MATCH_PROP_NAME);
	boolean match = new Boolean(matchString).booleanValue();

	String fileName = ((String)_getInputValue(FILENAME_NAME, "")).trim();
	String oldExt = ((String)_getInputValue(ARG1_NAME, "")).trim();
	String newExt = ((String)_getInputValue(ARG2_NAME, "")).trim();
	String replacement = "";

	if(fileName.endsWith(oldExt)||fileName.endsWith(oldExt.toUpperCase())){
	  replacement = 
	    fileName.substring(0, fileName.length() -oldExt.length());
	  if(match){
	    if(fileName.endsWith(oldExt.toUpperCase())){
	      replacement += newExt.toUpperCase();
	    }else{
	      replacement += newExt.toLowerCase();
	    }
	  }else{
	    replacement += newExt;
	  }
	}else{
	  if(newExt.startsWith(".")){
	    replacement = fileName +newExt;
	  }else{
	    replacement = fileName + "." + newExt;
	  }
	}
	
	_setOutputValue(REP_STRING_NAME, replacement);	
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = MATCH_PROP_NAME;
      String desc = MATCH_PROP_DESC;
      Object value = TRUE;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(TRUE);
      ops.add(FALSE);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable case matching";
      String REST_DESC = "Allowable boolean value for case matching";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
