/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector; 

/**
 * Processor that extends or reduces the size of a string by taking a substring
 * or by padding it with extra characters.  The adjustment can be made to the 
 * front or the back.
 *
 * @version 20 June 2003
 */
public class StringPadder extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Pads a String with specified string.";

  /** Input Node name of the String to check. */
  private static String STRING_NAME = "String";

  /** Input Node description of the String to check. */
  private static String STRING_DESC = "String to pad.";

  /** Output Node name of the returned String. */
  private static String RETURNED_NAME = "Returned String";

  /** Output Node description of the returned String. */
  private static String RETURNED_DESC = "Padded String.";

  /** Name of the size property. */
  private static String SIZE_NAME = "Size";

  /** Description of the size property. */
  private static String SIZE_DESC = "Size of the output string";

  /** Name of the side property. */
  private static String SIDE_NAME = "Side";

  /** Description of the side property. */
  private static String SIDE_DESC = "Side of string to change";

  /** Name of the front property value. */
  private static String FRONT_NAME = "Front";

  /** Name of the side property value. */
  private static String BACK_NAME = "Back";

  /** Input Node Name of the padding property. */
  private static String PADDING_NAME = "Padding";

  /** Input Node Description of the side property. */
  private static String PADDING_DESC = "String to be used to pad input string";

  /**
   * Constructs a String Padder with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringPadder(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a String Padder using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringPadder(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = String.class;
      inputNode = new InputNode(PADDING_NAME, PADDING_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = String.class;
      OutputNode outputNode = new OutputNode(RETURNED_NAME, RETURNED_DESC,
					     type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the input string
	String string = (String)_getInputValue(STRING_NAME, "");	
	String padding = (String)_getInputValue(PADDING_NAME, " ");

	String side = (String)_getPropertyValue(SIDE_NAME);
	Integer size = new Integer((String)_getPropertyValue(SIZE_NAME));

	if(side.equals(FRONT_NAME)){
	  if(string.length() >= size.intValue()){
	    _setOutputValue(RETURNED_NAME, 
			    string.substring(string.length()-size.intValue()));
	  }else{
	    String temp = padding;
	    while(temp.length() < size.intValue() - string.length()){
	      temp += padding;
	    }
	    _setOutputValue(RETURNED_NAME, 
			    temp.substring(0, size.intValue()-string.length())+
			    string);
	  }
	}else{
	  if(string.length() >= size.intValue()){
	    _setOutputValue(RETURNED_NAME, 
			    string.substring(0, size.intValue()));
	  }else{
	    String temp = padding;
	    while(temp.length() < size.intValue() - string.length()){
	      temp += padding;
	    }
	    _setOutputValue(RETURNED_NAME, string +
			    temp.substring(0,size.intValue()-string.length()));
	  }
	} 
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[2];

      // Default substring
      String name = SIZE_NAME;
      String desc = SIZE_DESC;
      Object value = "0";
      properties[0] = new ProcessorProperty(name, desc, value);

      // Default true string
      name = SIDE_NAME;
      desc = SIDE_DESC;
      value = BACK_NAME;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(FRONT_NAME);
      ops.add(BACK_NAME);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable padding sides";
      String REST_DESC = "Allowable sides to pad for this processor";
      Restriction restriction = 
	new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}



      
